from __future__ import annotations

import re
from collections.abc import Iterator
from dataclasses import replace

import numpy as np

from omnichunk.engine.code_engine import CodeEngine
from omnichunk.engine.prose_engine import ProseEngine
from omnichunk.sizing.nws import preprocess_nws_cumsum
from omnichunk.types import ByteRange, Chunk, ChunkOptions, ContentType, LineRange
from omnichunk.util.text_index import TextIndex


class HybridEngine:
    def __init__(self) -> None:
        self._code_engine = CodeEngine()
        self._prose_engine = ProseEngine()

    def chunk(self, filepath: str, content: str, options: ChunkOptions) -> list[Chunk]:
        chunks = list(self._iter_chunks(filepath, content, options))
        return _finalize_chunk_indexes(chunks)

    def stream(self, filepath: str, content: str, options: ChunkOptions) -> Iterator[Chunk]:
        for idx, chunk in enumerate(self._iter_chunks(filepath, content, options)):
            yield _with_unknown_total(chunk, idx)

    def _iter_chunks(self, filepath: str, content: str, options: ChunkOptions) -> Iterator[Chunk]:
        if not content.strip():
            return

        precomputed_text_index = options._precomputed_text_index
        if isinstance(precomputed_text_index, TextIndex):
            text_index = precomputed_text_index
        else:
            text_index = TextIndex(content)

        segments = _split_hybrid_segments(content)
        chunk_index = 0

        for segment in segments:
            segment_start, segment_end, segment_kind = segment
            segment_text = content[segment_start:segment_end]
            if not segment_text.strip():
                continue

            sub_options = replace(options)
            sub_options.content_type = (
                ContentType.CODE if segment_kind == "code" else ContentType.PROSE
            )
            if segment_start == 0 and segment_end == len(content):
                sub_options._precomputed_text_index = text_index
                sub_options._precomputed_nws_cumsum = options._precomputed_nws_cumsum
            else:
                bs = text_index.byte_offset_for_char(segment_start)
                be = text_index.byte_offset_for_char(segment_end)
                pc = options._precomputed_nws_cumsum
                if isinstance(pc, np.ndarray) and int(pc.size) > be:
                    sub_options._precomputed_nws_cumsum = pc[bs : be + 1] - pc[bs]
                else:
                    sub_options._precomputed_nws_cumsum = preprocess_nws_cumsum(
                        segment_text,
                        backend=options.nws_backend,
                    )
                sub_options._precomputed_text_index = TextIndex.from_parent_slice(
                    text_index,
                    segment_start,
                    segment_end,
                )

            if segment_kind == "code":
                local_chunks = self._code_engine.stream(filepath, segment_text, sub_options)
            else:
                local_chunks = self._prose_engine.stream(filepath, segment_text, sub_options)

            for local_chunk in local_chunks:
                rebased = _rebase_chunk(text_index, segment_start, local_chunk)
                yield Chunk(
                    text=rebased.text,
                    contextualized_text=rebased.contextualized_text,
                    byte_range=rebased.byte_range,
                    line_range=rebased.line_range,
                    index=chunk_index,
                    total_chunks=-1,
                    context=rebased.context,
                    token_count=rebased.token_count,
                    char_count=rebased.char_count,
                    nws_count=rebased.nws_count,
                )
                chunk_index += 1


def _split_hybrid_segments(content: str) -> list[tuple[int, int, str]]:
    if "# %%" in content:
        return _split_by_cells(content)

    docstring_matches = list(re.finditer(r'"""[\s\S]*?"""|\'\'\'[\s\S]*?\'\'\'', content))
    doc_chars = sum(m.end() - m.start() for m in docstring_matches)
    if doc_chars / max(1, len(content)) <= 0.4:
        return [(0, len(content), "code")]

    segments: list[tuple[int, int, str]] = []
    cursor = 0
    for match in docstring_matches:
        if match.start() > cursor:
            segments.append((cursor, match.start(), "code"))
        segments.append((match.start(), match.end(), "prose"))
        cursor = match.end()
    if cursor < len(content):
        segments.append((cursor, len(content), "code"))
    return segments


def _split_by_cells(content: str) -> list[tuple[int, int, str]]:
    boundaries = [m.start() for m in re.finditer(r"(?m)^#\s*%%", content)]
    if not boundaries:
        return [(0, len(content), "code")]

    boundaries = [0] + boundaries + [len(content)]
    segments: list[tuple[int, int, str]] = []
    for idx in range(len(boundaries) - 1):
        start = boundaries[idx]
        end = boundaries[idx + 1]
        text = content[start:end]
        segment_type = "prose" if _looks_markdown_cell(text) else "code"
        segments.append((start, end, segment_type))
    return segments


def _looks_markdown_cell(text: str) -> bool:
    if "# %% [markdown]" in text.lower():
        return True
    stripped = text.strip()
    return stripped.startswith('"""') or stripped.startswith("'''")


def _rebase_chunk(text_index: TextIndex, segment_char_start: int, chunk: Chunk) -> Chunk:
    segment_byte_start = text_index.byte_offset_for_char(segment_char_start)
    line_offset = text_index.line_for_char(segment_char_start)

    return Chunk(
        text=chunk.text,
        contextualized_text=chunk.contextualized_text,
        byte_range=ByteRange(
            start=segment_byte_start + chunk.byte_range.start,
            end=segment_byte_start + chunk.byte_range.end,
        ),
        line_range=LineRange(
            start=line_offset + chunk.line_range.start,
            end=line_offset + chunk.line_range.end,
        ),
        index=chunk.index,
        total_chunks=chunk.total_chunks,
        context=chunk.context,
        token_count=chunk.token_count,
        char_count=chunk.char_count,
        nws_count=chunk.nws_count,
    )


def _finalize_chunk_indexes(chunks: list[Chunk]) -> list[Chunk]:
    total = len(chunks)
    return [
        Chunk(
            text=chunk.text,
            contextualized_text=chunk.contextualized_text,
            byte_range=chunk.byte_range,
            line_range=chunk.line_range,
            index=idx,
            total_chunks=total,
            context=chunk.context,
            token_count=chunk.token_count,
            char_count=chunk.char_count,
            nws_count=chunk.nws_count,
        )
        for idx, chunk in enumerate(chunks)
    ]


def _with_unknown_total(chunk: Chunk, index: int) -> Chunk:
    return Chunk(
        text=chunk.text,
        contextualized_text=chunk.contextualized_text,
        byte_range=chunk.byte_range,
        line_range=chunk.line_range,
        index=index,
        total_chunks=-1,
        context=chunk.context,
        token_count=chunk.token_count,
        char_count=chunk.char_count,
        nws_count=chunk.nws_count,
    )



from __future__ import annotations

import re

from omnichunk.formats.types import FormatSegment, LoadedDocument

_CODE_ENV_PATTERN = re.compile(
    r"\\begin\{(lstlisting|verbatim|minted)\}"
    r"(?:\[[^\]]*\])?"
    r"([\s\S]*?)"
    r"\\end\{\1\}",
)

_SECTION_PATTERN = re.compile(
    r"\\(?:part|chapter|section|subsection|subsubsection)\*?\{([^}]*)\}",
)


def load_latex(content: str) -> LoadedDocument:
    """Split LaTeX into prose and code (verbatim-like) segments."""
    warnings: list[str] = []
    text = content
    segments: list[FormatSegment] = []

    if _unbalanced_begin_end(text):
        warnings.append("possible_unbalanced_begin_end")

    last = 0
    for match in _CODE_ENV_PATTERN.finditer(text):
        if match.start() > last:
            prose = text[last : match.start()]
            _append_tex_prose_segments(prose, last, segments, text)
        env = match.group(1)
        start = match.start()
        end = match.end()
        segments.append(
            FormatSegment(
                char_start=start,
                char_end=end,
                kind="code",
                metadata={"latex_env": env},
            )
        )
        last = end

    if last < len(text):
        prose = text[last:]
        _append_tex_prose_segments(prose, last, segments, text)

    ordered = sorted(segments, key=lambda s: s.char_start)
    filtered = tuple(
        s
        for s in ordered
        if s.char_end > s.char_start and text[s.char_start : s.char_end].strip()
    )

    return LoadedDocument(
        text=text,
        segments=filtered,
        format_name="latex",
        warnings=tuple(warnings),
    )


def _unbalanced_begin_end(text: str) -> bool:
    begins = len(re.findall(r"\\begin\{", text))
    ends = len(re.findall(r"\\end\{", text))
    return begins != ends


def _append_tex_prose_segments(
    prose: str,
    base_offset: int,
    segments: list[FormatSegment],
    full_text: str,
) -> None:
    """Split prose regions by section commands for sharper chunk boundaries."""
    if not prose:
        return

    sec_iter = list(_SECTION_PATTERN.finditer(prose))
    if not sec_iter:
        start = base_offset
        end = base_offset + len(prose)
        if full_text[start:end].strip():
            segments.append(
                FormatSegment(
                    char_start=start,
                    char_end=end,
                    kind="prose",
                    metadata={"latex": "body"},
                )
            )
        return

    pos = 0
    for m in sec_iter:
        if m.start() > pos:
            chunk = prose[pos : m.start()]
            if chunk.strip():
                segments.append(
                    FormatSegment(
                        char_start=base_offset + pos,
                        char_end=base_offset + m.start(),
                        kind="prose",
                        metadata={"latex": "preamble_segment"},
                    )
                )
        title = m.group(1)
        segments.append(
            FormatSegment(
                char_start=base_offset + m.start(),
                char_end=base_offset + m.end(),
                kind="prose",
                metadata={"latex": "section_heading", "title": title},
            )
        )
        pos = m.end()

    if pos < len(prose):
        tail = prose[pos:]
        if tail.strip():
            segments.append(
                FormatSegment(
                    char_start=base_offset + pos,
                    char_end=base_offset + len(prose),
                    kind="prose",
                    metadata={"latex": "post_section"},
                )
            )




from __future__ import annotations

from omnichunk import Chunker, build_chunk_graph
from omnichunk.graph import ChunkGraph


def test_build_chunk_graph_basic() -> None:
    code = (
        "def alpha():\n    return 1\n\n"
        "def beta():\n    alpha()\n\n"
        "def gamma():\n    alpha()\n"
    )
    chunker = Chunker(max_chunk_size=64, size_unit="chars")
    chunks = chunker.chunk("m.py", code)
    graph = build_chunk_graph(chunks, min_entity_occurrences=1)
    assert isinstance(graph, ChunkGraph)
    assert graph.chunk_count == len(chunks)
    assert graph.nodes


def test_build_chunk_graph_cross_chunk_entity() -> None:
    code = "def shared():\n    pass\n" + "\n" * 3 + "def caller():\n    shared()\n"
    chunker = Chunker(max_chunk_size=40, size_unit="chars")
    chunks = chunker.chunk("m.py", code)
    graph = build_chunk_graph(chunks, min_entity_occurrences=1)
    all_chunk_lists = [list(n.chunk_indices) for n in graph.nodes.values()]
    multi_chunk_entities = [cl for cl in all_chunk_lists if len(cl) >= 2]
    if multi_chunk_entities:
        assert graph.edges


def test_chunk_neighbors_symmetric() -> None:
    code = "def f():\n    pass\n" * 6
    chunker = Chunker(max_chunk_size=48, size_unit="chars")
    chunks = chunker.chunk("m.py", code)
    graph = build_chunk_graph(chunks, min_entity_occurrences=1)
    for edge in graph.edges:
        assert edge.chunk_b in graph.chunk_neighbors(edge.chunk_a)
        assert edge.chunk_a in graph.chunk_neighbors(edge.chunk_b)


def test_chunk_graph_to_dict_round_trip() -> None:
    code = "def foo():\n    pass\n" * 3
    chunker = Chunker(max_chunk_size=40, size_unit="chars")
    chunks = chunker.chunk("x.py", code)
    graph = build_chunk_graph(chunks, min_entity_occurrences=1)
    data = graph.to_dict()
    restored = ChunkGraph.from_dict(data)
    assert restored.chunk_count == graph.chunk_count
    assert set(restored.nodes) == set(graph.nodes)


def test_build_chunk_graph_empty() -> None:
    graph = build_chunk_graph([])
    assert graph.chunk_count == 0
    assert not graph.nodes
    assert not graph.edges


def test_entity_chunks_unknown_entity_returns_empty() -> None:
    graph = build_chunk_graph([])
    assert graph.entity_chunks("nonexistent") == []


def test_chunk_neighbors_no_edges_returns_empty() -> None:
    graph = build_chunk_graph([])
    assert graph.chunk_neighbors(0) == []


def test_build_chunk_graph_min_occurrences_filters() -> None:
    code = "def unique_only():\n    pass\n"
    chunker = Chunker(max_chunk_size=40, size_unit="chars")
    chunks = chunker.chunk("x.py", code)
    graph = build_chunk_graph(chunks, min_entity_occurrences=2)
    assert "unique_only" not in graph.nodes



from __future__ import annotations

import json
from typing import Any

from omnichunk.formats.types import FormatSegment, LoadedDocument


def load_ipynb(content: str) -> LoadedDocument:
    """Parse a Jupyter notebook JSON string into canonical text and segments."""
    warnings: list[str] = []
    try:
        data: dict[str, Any] = json.loads(content)
    except json.JSONDecodeError as exc:
        return LoadedDocument(
            text="",
            segments=(),
            format_name="ipynb",
            warnings=(f"invalid_json: {exc}",),
        )

    cells = data.get("cells")
    if not isinstance(cells, list):
        return LoadedDocument(
            text="",
            segments=(),
            format_name="ipynb",
            warnings=("missing_or_invalid_cells",),
        )

    parts: list[str] = []
    segments: list[FormatSegment] = []
    cursor = 0

    for idx, cell in enumerate(cells):
        if not isinstance(cell, dict):
            warnings.append(f"cell_{idx}_not_object")
            continue

        ctype = str(cell.get("cell_type", "") or "")
        src = cell.get("source", "")
        cell_text = "".join(str(x) for x in src) if isinstance(src, list) else str(src)

        if ctype == "markdown":
            if cell_text:
                start = cursor
                parts.append(cell_text)
                cursor += len(cell_text)
                segments.append(
                    FormatSegment(
                        char_start=start,
                        char_end=cursor,
                        kind="prose",
                        metadata={"cell": idx, "cell_type": "markdown"},
                    )
                )
        elif ctype == "code":
            lang = _code_language(cell)
            if cell_text:
                start = cursor
                parts.append(cell_text)
                cursor += len(cell_text)
                segments.append(
                    FormatSegment(
                        char_start=start,
                        char_end=cursor,
                        kind="code",
                        metadata={"cell": idx, "cell_type": "code", "language": lang},
                    )
                )
            out_text = _extract_cell_outputs(cell)
            if out_text.strip():
                start = cursor
                parts.append(out_text)
                cursor += len(out_text)
                segments.append(
                    FormatSegment(
                        char_start=start,
                        char_end=cursor,
                        kind="prose",
                        metadata={"cell": idx, "cell_type": "output"},
                    )
                )
        elif ctype == "raw":
            if cell_text:
                start = cursor
                parts.append(cell_text)
                cursor += len(cell_text)
                segments.append(
                    FormatSegment(
                        char_start=start,
                        char_end=cursor,
                        kind="prose",
                        metadata={"cell": idx, "cell_type": "raw"},
                    )
                )
        else:
            warnings.append(f"unknown_cell_type:{ctype}")

    full_text = "".join(parts)
    return LoadedDocument(
        text=full_text,
        segments=tuple(segments),
        format_name="ipynb",
        warnings=tuple(warnings),
    )


def _code_language(cell: dict[str, Any]) -> str:
    meta = cell.get("metadata")
    if isinstance(meta, dict):
        lang = meta.get("language")
        if isinstance(lang, str) and lang:
            return lang
    return "python"


def _extract_cell_outputs(cell: dict[str, Any]) -> str:
    outputs = cell.get("outputs")
    if not isinstance(outputs, list) or not outputs:
        return ""

    lines: list[str] = []
    for out in outputs:
        if not isinstance(out, dict):
            continue
        otype = str(out.get("output_type", "") or "")
        if otype == "stream":
            text = out.get("text", "")
            if isinstance(text, list):
                lines.append("".join(str(x) for x in text))
            else:
                lines.append(str(text))
        elif otype in ("display_data", "execute_result"):
            data = out.get("data")
            if isinstance(data, dict):
                plain = data.get("text/plain")
                if isinstance(plain, list):
                    lines.append("".join(str(x) for x in plain))
                elif isinstance(plain, str):
                    lines.append(plain)
        elif otype == "error":
            tb = out.get("traceback")
            if isinstance(tb, list):
                lines.append("\n".join(str(x) for x in tb))
        # Skip binary image/svg-heavy payloads deterministically

    if not lines:
        return ""

    return "[output]\n" + "\n".join(lines) + "\n"



from __future__ import annotations

from omnichunk.quality import compute_chunk_quality_scores, compute_chunk_stats
from omnichunk.types import (
    ByteRange,
    Chunk,
    ChunkContext,
    EntityInfo,
    EntityType,
    LineRange,
)


def test_quality_scores_penalize_partial_entities_and_size_extremes() -> None:
    clean_chunk = Chunk(
        text="def clean():\n    return 1\n",
        contextualized_text="def clean():\n    return 1\n",
        byte_range=ByteRange(0, 24),
        line_range=LineRange(0, 1),
        index=0,
        total_chunks=2,
        context=ChunkContext(
            filepath="sample.py",
            language="python",
            entities=[
                EntityInfo(
                    name="clean",
                    type=EntityType.FUNCTION,
                    byte_range=ByteRange(0, 24),
                    line_range=LineRange(0, 1),
                )
            ],
            breadcrumb=["clean"],
        ),
        token_count=7,
        char_count=24,
        nws_count=19,
    )

    partial_chunk = Chunk(
        text="return a + b",
        contextualized_text="return a + b",
        byte_range=ByteRange(24, 36),
        line_range=LineRange(2, 2),
        index=1,
        total_chunks=2,
        context=ChunkContext(
            filepath="sample.py",
            language="python",
            entities=[
                EntityInfo(
                    name="calc",
                    type=EntityType.FUNCTION,
                    byte_range=ByteRange(10, 48),
                    line_range=LineRange(0, 3),
                    is_partial=True,
                )
            ],
            breadcrumb=["calc", "nested"],
        ),
        token_count=4,
        char_count=12,
        nws_count=9,
    )

    scores = compute_chunk_quality_scores(
        [clean_chunk, partial_chunk],
        min_chunk_size=12,
        max_chunk_size=36,
        size_unit="chars",
    )

    assert len(scores) == 2
    assert scores[0].overall > scores[1].overall
    assert scores[0].entity_integrity == 1.0
    assert scores[1].entity_integrity < 1.0


def test_chunk_stats_aggregate_size_and_entity_distribution() -> None:
    chunks = [
        Chunk(
            text="a",
            contextualized_text="a",
            byte_range=ByteRange(0, 1),
            line_range=LineRange(0, 0),
            index=0,
            total_chunks=2,
            context=ChunkContext(
                filepath="x.py",
                language="python",
                entities=[EntityInfo(name="x", type=EntityType.FUNCTION)],
            ),
            token_count=1,
            char_count=1,
            nws_count=1,
        ),
        Chunk(
            text="bc",
            contextualized_text="bc",
            byte_range=ByteRange(1, 3),
            line_range=LineRange(1, 1),
            index=1,
            total_chunks=2,
            context=ChunkContext(
                filepath="x.py",
                language="python",
                entities=[EntityInfo(name="Y", type=EntityType.CLASS)],
            ),
            token_count=2,
            char_count=2,
            nws_count=2,
        ),
    ]

    stats = compute_chunk_stats(chunks, size_unit="chars")

    assert stats.total_chunks == 2
    assert stats.average_size == 1.5
    assert stats.min_size == 1
    assert stats.max_size == 2
    assert stats.entity_distribution == {"class": 1, "function": 1}



from __future__ import annotations

from pathlib import Path

from omnichunk.serialization import stable_chunk_id
from omnichunk.store import ChunkStore


def test_chunk_store_index_and_query(tmp_path: Path) -> None:
    d = tmp_path / "proj"
    d.mkdir()
    (d / "a.py").write_text("def f():\n    return 1\n", encoding="utf-8")

    db = tmp_path / "store.db"
    store = ChunkStore(db)
    try:
        r = store.index(str(d), glob="**/*.py")
        assert r.files_scanned == 1
        assert r.files_updated == 1
        assert r.files_skipped == 0

        chunks = store.query(str(d / "a.py"))
        assert chunks
        assert all("def f" in c.text for c in chunks)
    finally:
        store.close()


def test_chunk_store_sync_skips_unchanged(tmp_path: Path) -> None:
    d = tmp_path / "proj"
    d.mkdir()
    f = d / "x.py"
    f.write_text("x = 1\n", encoding="utf-8")

    db = tmp_path / "store.db"
    store = ChunkStore(db)
    try:
        r1 = store.sync(str(d), glob="**/*.py")
        assert r1.files_updated == 1
        assert r1.files_skipped == 0

        r2 = store.sync(str(d), glob="**/*.py")
        assert r2.files_updated == 0
        assert r2.files_skipped == 1
    finally:
        store.close()


def test_chunk_store_sync_updates_on_change(tmp_path: Path) -> None:
    d = tmp_path / "proj"
    d.mkdir()
    f = d / "m.py"
    f.write_text("a = 1\n", encoding="utf-8")

    db = tmp_path / "store.db"
    store = ChunkStore(db)
    try:
        store.sync(str(d), glob="**/*.py")
        ids_v1 = {stable_chunk_id(c) for c in store.query(str(f))}

        f.write_text("a = 1\nb = 2\n", encoding="utf-8")
        r = store.sync(str(d), glob="**/*.py")
        assert r.files_updated == 1
        key = str(f.resolve())
        assert key in r.diffs
        assert r.diffs[key].removed_ids
        ids_v2 = {stable_chunk_id(c) for c in store.query(str(f))}
        assert ids_v1.isdisjoint(ids_v2)
    finally:
        store.close()


def test_chunk_store_removes_deleted_file(tmp_path: Path) -> None:
    d = tmp_path / "proj"
    d.mkdir()
    f = d / "gone.py"
    f.write_text("print(1)\n", encoding="utf-8")

    db = tmp_path / "store.db"
    store = ChunkStore(db)
    try:
        store.sync(str(d), glob="**/*.py")
        ids_before = {stable_chunk_id(c) for c in store.query(str(f))}
        assert ids_before

        f.unlink()
        r = store.sync(str(d), glob="**/*.py")
        assert r.files_deleted >= 1
        assert set(r.removed_chunk_ids) >= ids_before
        # query path may still exist as string - file not on disk
        assert store.query(str(d / "gone.py")) == []
    finally:
        store.close()


def test_chunk_store_foreign_keys_cascade(tmp_path: Path) -> None:
    db = tmp_path / "store.db"
    store = ChunkStore(db)
    try:
        store._conn.execute(
            "INSERT INTO files (path, mtime_ns, size, content_sha256, last_indexed_at) "
            "VALUES (?, 0, 0, 'x', 0.0)",
            ("/tmp/fake",),
        )
        store._conn.execute(
            "INSERT INTO chunks (stable_id, filepath, chunk_json) VALUES (?, ?, ?)",
            ("sid1", "/tmp/fake", "{}"),
        )
        store._conn.execute("DELETE FROM files WHERE path = ?", ("/tmp/fake",))
        cur = store._conn.execute(
            "SELECT COUNT(*) FROM chunks WHERE stable_id = ?",
            ("sid1",),
        )
        n = cur.fetchone()[0]
        assert n == 0
    finally:
        store.close()


def test_chunk_store_empty_glob(tmp_path: Path) -> None:
    d = tmp_path / "empty"
    d.mkdir()
    db = tmp_path / "store.db"
    store = ChunkStore(db)
    try:
        r = store.sync(str(d), glob="**/*.py")
        assert r.files_scanned == 0
    finally:
        store.close()



from __future__ import annotations

import re
from collections.abc import Callable, Iterator
from dataclasses import replace
from typing import Any

from omnichunk.context.format import format_contextualized_text
from omnichunk.engine.code_engine import CodeEngine
from omnichunk.engine.markup_engine import MarkupEngine
from omnichunk.parser.markdown_parser import ProseNode, parse_markdown
from omnichunk.sizing.counter import make_size_counter, make_token_counter
from omnichunk.sizing.nws import get_nws_count, preprocess_nws_cumsum
from omnichunk.types import (
    ByteRange,
    Chunk,
    ChunkContext,
    ChunkOptions,
    ContentType,
    EntityInfo,
    EntityType,
    Language,
    LineRange,
)
from omnichunk.util.detect import detect_language
from omnichunk.util.text_index import TextIndex
from omnichunk.windowing.overlap import build_line_overlap_text

_SPLIT_HIERARCHY: list[str] = [
    r"(?:\r?\n){2,}",
    r"\t+",
    r"\s+",
    r"[\.\?!\*]+",
    r"[;,\(\)\[\]\"'`]+",
    r"[:—…]+",
    r"[/\\–&-]+",
    r"",
]

_CODE_FENCE_LANGUAGE_ALIASES: dict[str, str] = {
    "py": "python",
    "python": "python",
    "js": "javascript",
    "javascript": "javascript",
    "ts": "typescript",
    "typescript": "typescript",
    "tsx": "typescript",
    "rs": "rust",
    "rust": "rust",
    "go": "go",
    "java": "java",
    "c": "c",
    "cpp": "cpp",
    "c++": "cpp",
    "cs": "csharp",
    "csharp": "csharp",
    "rb": "ruby",
    "ruby": "ruby",
    "php": "php",
    "kt": "kotlin",
    "kotlin": "kotlin",
    "swift": "swift",
}

_CODE_LANGUAGES: set[str] = {
    "python",
    "javascript",
    "typescript",
    "rust",
    "go",
    "java",
    "c",
    "cpp",
    "csharp",
    "ruby",
    "php",
    "kotlin",
    "swift",
    "scala",
    "haskell",
    "lua",
    "zig",
    "elixir",
}

_MARKUP_FENCE_LANGUAGE_ALIASES: dict[str, str] = {
    "json": "json",
    "yaml": "yaml",
    "yml": "yaml",
    "toml": "toml",
    "html": "html",
    "xml": "xml",
}

_MARKUP_LANGUAGES: set[str] = {"json", "yaml", "toml", "html", "xml"}


class ProseEngine:
    def __init__(self) -> None:
        self._code_engine = CodeEngine()
        self._markup_engine = MarkupEngine()

    def chunk(self, filepath: str, content: str, options: ChunkOptions) -> list[Chunk]:
        chunks = list(self._iter_chunks(filepath, content, options))
        return _finalize_chunk_indexes(chunks)

    def stream(self, filepath: str, content: str, options: ChunkOptions) -> Iterator[Chunk]:
        for idx, chunk in enumerate(self._iter_chunks(filepath, content, options)):
            yield _with_unknown_total(chunk, idx)

    def _iter_chunks(self, filepath: str, content: str, options: ChunkOptions) -> Iterator[Chunk]:
        if not content.strip():
            return

        precomputed_text_index = options._precomputed_text_index
        if isinstance(precomputed_text_index, TextIndex):
            text_index = precomputed_text_index
        else:
            text_index = TextIndex(content)

        language = options.language or detect_language(filepath=filepath, content=content)
        size_counter = make_size_counter(
            options.size_unit,
            tokenizer=options.tokenizer,
            chunk_size=options.max_chunk_size,
        )
        token_counter = make_token_counter(options.tokenizer, chunk_size=options.max_chunk_size)

        if language == "markdown":
            _, nodes = parse_markdown(content)
        else:
            nodes = _plaintext_nodes(content, text_index)

        windows = _build_node_windows(
            nodes,
            max_size=options.max_chunk_size,
            min_size=options.min_chunk_size,
            size_counter=size_counter,
            content=content,
            preserve_code_blocks=(language == "markdown"),
        )

        if not windows:
            windows = [[(0, len(content), [], "paragraph")]]

        ranges = _windows_to_contiguous_ranges(windows, len(content))
        ranges = _filter_empty_ranges(content, ranges)

        cumsum = options._precomputed_nws_cumsum
        if cumsum is None:
            cumsum = preprocess_nws_cumsum(content, backend=options.nws_backend)

        previous_text = ""
        chunk_index = 0

        for start, end, hierarchy, section_type in ranges:
            if end <= start:
                continue

            text = content[start:end]
            if not text.strip():
                continue

            if language == "markdown" and section_type == EntityType.CODE_BLOCK.value:
                code_chunks = self._chunk_markdown_code_block(
                    filepath=filepath,
                    block_text=text,
                    block_char_start=start,
                    heading_hierarchy=hierarchy,
                    options=options,
                    text_index=text_index,
                    cumsum=cumsum,
                    start_index=chunk_index,
                )
                if code_chunks:
                    for chunk in code_chunks:
                        yield chunk
                        previous_text = chunk.text
                        chunk_index += 1
                    continue

            line_start = text_index.line_for_char(start)
            line_end = text_index.line_for_char(max(start, end - 1))
            context = ChunkContext(
                filepath=filepath,
                language=language,
                content_type=ContentType.PROSE,
                heading_hierarchy=hierarchy,
                section_type=section_type,
                entities=_entities_for_section(
                    section_type, text, start, end, line_start, line_end
                ),
            )

            overlap_text = build_line_overlap_text(previous_text, options.overlap_lines)
            contextualized_text = (
                text
                if options.context_mode == "none"
                else format_contextualized_text(text, context=context, overlap_text=overlap_text)
            )

            byte_start = text_index.byte_offset_for_char(start)
            byte_end = text_index.byte_offset_for_char(end)

            yield Chunk(
                text=text,
                contextualized_text=contextualized_text,
                byte_range=ByteRange(byte_start, byte_end),
                line_range=LineRange(line_start, max(line_start, line_end)),
                index=chunk_index,
                total_chunks=-1,
                context=context,
                token_count=token_counter(text),
                char_count=len(text),
                nws_count=get_nws_count(cumsum, byte_start, byte_end),
            )
            previous_text = text
            chunk_index += 1

    def _chunk_markdown_code_block(
        self,
        *,
        filepath: str,
        block_text: str,
        block_char_start: int,
        heading_hierarchy: list[str],
        options: ChunkOptions,
        text_index: TextIndex,
        cumsum: Any,
        start_index: int,
    ) -> list[Chunk]:
        parts = _split_fenced_code_block(block_text)
        if parts is None:
            return []

        prefix_text, code_text, suffix_text, fence_language = parts
        token_counter = make_token_counter(options.tokenizer, chunk_size=options.max_chunk_size)
        chunks: list[Chunk] = []
        delegated_text_index = TextIndex(code_text) if code_text else None
        delegated_cumsum = (
            preprocess_nws_cumsum(code_text, backend=options.nws_backend) if code_text else None
        )

        cursor = block_char_start

        if prefix_text and prefix_text.strip():
            prefix_start = cursor
            prefix_end = prefix_start + len(prefix_text)
            chunks.append(
                _build_markdown_fence_chunk(
                    filepath=filepath,
                    language=options.language or "markdown",
                    text=prefix_text,
                    heading_hierarchy=heading_hierarchy,
                    start_char=prefix_start,
                    end_char=prefix_end,
                    index=start_index + len(chunks),
                    options=options,
                    text_index=text_index,
                    cumsum=cumsum,
                    token_counter=token_counter,
                )
            )
        cursor += len(prefix_text)

        if code_text:
            resolved_language = _resolve_code_fence_language(fence_language, code_text)
            if resolved_language is not None:
                sub_options = replace(
                    options,
                    content_type=ContentType.CODE,
                    overlap_lines=0,
                    _precomputed_text_index=delegated_text_index,
                    _precomputed_nws_cumsum=delegated_cumsum,
                )
                sub_options.language = resolved_language  # type: ignore[assignment]

                code_chunks = list(self._code_engine.stream(filepath, code_text, sub_options))
                delegated = _rebase_delegated_fence_chunks(
                    delegated_chunks=code_chunks,
                    heading_hierarchy=heading_hierarchy,
                    options=options,
                    text_index=text_index,
                    byte_offset=text_index.byte_offset_for_char(cursor),
                    cumsum=cumsum,
                    start_index=start_index + len(chunks),
                )
                if delegated:
                    chunks.extend(delegated)
                elif code_text.strip():
                    code_start = cursor
                    code_end = code_start + len(code_text)
                    chunks.append(
                        _build_markdown_fence_chunk(
                            filepath=filepath,
                            language=options.language or "markdown",
                            text=code_text,
                            heading_hierarchy=heading_hierarchy,
                            start_char=code_start,
                            end_char=code_end,
                            index=start_index + len(chunks),
                            options=options,
                            text_index=text_index,
                            cumsum=cumsum,
                            token_counter=token_counter,
                        )
                    )
            else:
                resolved_markup_language = _resolve_markup_fence_language(fence_language, code_text)
                if resolved_markup_language is not None:
                    sub_options = replace(
                        options,
                        content_type=ContentType.MARKUP,
                        overlap_lines=0,
                        _precomputed_text_index=delegated_text_index,
                        _precomputed_nws_cumsum=delegated_cumsum,
                    )
                    sub_options.language = resolved_markup_language  # type: ignore[assignment]

                    markup_chunks = list(
                        self._markup_engine.stream(filepath, code_text, sub_options)
                    )
                    delegated = _rebase_delegated_fence_chunks(
                        delegated_chunks=markup_chunks,
                        heading_hierarchy=heading_hierarchy,
                        options=options,
                        text_index=text_index,
                        byte_offset=text_index.byte_offset_for_char(cursor),
                        cumsum=cumsum,
                        start_index=start_index + len(chunks),
                    )
                    if delegated:
                        chunks.extend(delegated)
                    elif code_text.strip():
                        code_start = cursor
                        code_end = code_start + len(code_text)
                        chunks.append(
                            _build_markdown_fence_chunk(
                                filepath=filepath,
                                language=options.language or "markdown",
                                text=code_text,
                                heading_hierarchy=heading_hierarchy,
                                start_char=code_start,
                                end_char=code_end,
                                index=start_index + len(chunks),
                                options=options,
                                text_index=text_index,
                                cumsum=cumsum,
                                token_counter=token_counter,
                            )
                        )
                elif code_text.strip():
                    code_start = cursor
                    code_end = code_start + len(code_text)
                    chunks.append(
                        _build_markdown_fence_chunk(
                            filepath=filepath,
                            language=options.language or "markdown",
                            text=code_text,
                            heading_hierarchy=heading_hierarchy,
                            start_char=code_start,
                            end_char=code_end,
                            index=start_index + len(chunks),
                            options=options,
                            text_index=text_index,
                            cumsum=cumsum,
                            token_counter=token_counter,
                        )
                    )
        cursor += len(code_text)

        if suffix_text and suffix_text.strip():
            suffix_start = cursor
            suffix_end = suffix_start + len(suffix_text)
            chunks.append(
                _build_markdown_fence_chunk(
                    filepath=filepath,
                    language=options.language or "markdown",
                    text=suffix_text,
                    heading_hierarchy=heading_hierarchy,
                    start_char=suffix_start,
                    end_char=suffix_end,
                    index=start_index + len(chunks),
                    options=options,
                    text_index=text_index,
                    cumsum=cumsum,
                    token_counter=token_counter,
                )
            )

        return chunks


def _plaintext_nodes(content: str, text_index: TextIndex | None = None) -> list[ProseNode]:
    nodes: list[ProseNode] = []
    index = text_index or TextIndex(content)
    cursor = 0
    parts = re.split(r"(\n\s*\n)", content)
    for part in parts:
        if not part:
            continue
        start = cursor
        end = cursor + len(part)
        cursor = end
        if not part.strip():
            continue
        line_start = index.line_for_char(start)
        line_end = index.line_for_char(max(start, end - 1))
        nodes.append(
            ProseNode(
                kind=EntityType.PARAGRAPH,
                text=part,
                byte_range=ByteRange(start, end),
                line_range=LineRange(line_start, max(line_start, line_end)),
                heading_hierarchy=[],
            )
        )
    return nodes


def _build_node_windows(
    nodes: list[ProseNode],
    *,
    max_size: int,
    min_size: int,
    size_counter: Callable[[str], int],
    content: str,
    preserve_code_blocks: bool = False,
) -> list[list[tuple[int, int, list[str], str]]]:
    windows: list[list[tuple[int, int, list[str], str]]] = []
    current: list[tuple[int, int, list[str], str]] = []
    current_size = 0

    for node in nodes:
        node_size = size_counter(node.text)
        node_payload = (
            node.byte_range.start,
            node.byte_range.end,
            list(node.heading_hierarchy),
            node.kind.value,
        )

        if preserve_code_blocks and node.kind == EntityType.CODE_BLOCK:
            if current:
                windows.append(current)
                current = []
                current_size = 0
            windows.append([node_payload])
            continue

        if node_size <= max_size and current_size + node_size <= max_size:
            current.append(node_payload)
            current_size += node_size
            continue

        if node_size > max_size:
            if current:
                windows.append(current)
                current = []
                current_size = 0

            split_ranges = _split_oversized_text(
                node.text,
                base_start=node.byte_range.start,
                max_size=max_size,
                size_counter=size_counter,
            )
            windows.extend(
                [
                    [(start, end, list(node.heading_hierarchy), node.kind.value)]
                    for start, end in split_ranges
                    if end > start
                ]
            )
            continue

        if current:
            windows.append(current)
        current = [node_payload]
        current_size = node_size

    if current:
        windows.append(current)

    if len(windows) <= 1:
        return windows

    idx = 0
    while idx < len(windows):
        window = windows[idx]
        if preserve_code_blocks and _window_has_section_type(window, EntityType.CODE_BLOCK.value):
            idx += 1
            continue

        size = size_counter(content[window[0][0] : window[-1][1]])
        if size >= min_size:
            idx += 1
            continue

        if idx + 1 < len(windows):
            if preserve_code_blocks and _window_has_section_type(
                windows[idx + 1], EntityType.CODE_BLOCK.value
            ):
                idx += 1
                continue
            windows[idx].extend(windows[idx + 1])
            del windows[idx + 1]
            continue

        if idx > 0:
            if preserve_code_blocks and _window_has_section_type(
                windows[idx - 1], EntityType.CODE_BLOCK.value
            ):
                idx += 1
                continue
            windows[idx - 1].extend(windows[idx])
            del windows[idx]
            idx -= 1
            continue

        idx += 1

    return windows


def _window_has_section_type(
    window: list[tuple[int, int, list[str], str]], section_type: str
) -> bool:
    return any(item[3] == section_type for item in window)


def _split_oversized_text(
    text: str,
    *,
    base_start: int,
    max_size: int,
    size_counter: Callable[[str], int],
) -> list[tuple[int, int]]:
    if size_counter(text) <= max_size:
        return [(base_start, base_start + len(text))]

    splits = _hierarchical_split_points(text)
    if not splits:
        splits = list(range(1, len(text)))

    chunks: list[tuple[int, int]] = []
    cursor = 0
    while cursor < len(text):
        best_end = cursor + 1
        for point in splits:
            if point <= cursor:
                continue
            candidate = text[cursor:point]
            if size_counter(candidate) <= max_size:
                best_end = point
            else:
                break

        if best_end <= cursor:
            best_end = min(len(text), cursor + max_size)

        chunks.append((base_start + cursor, base_start + best_end))
        cursor = best_end

    return chunks


def _hierarchical_split_points(text: str) -> list[int]:
    points: set[int] = set()
    for pattern in _SPLIT_HIERARCHY:
        if not pattern:
            continue
        for match in re.finditer(pattern, text):
            points.add(match.end())
    points.add(len(text))
    return sorted(p for p in points if p > 0)


def _windows_to_contiguous_ranges(
    windows: list[list[tuple[int, int, list[str], str]]],
    total_len: int,
) -> list[tuple[int, int, list[str], str]]:
    if not windows:
        return []

    windows = sorted(windows, key=lambda w: w[0][0])
    ranges: list[tuple[int, int, list[str], str]] = []
    cursor = 0

    for window in windows:
        window_end = max(item[1] for item in window)
        hierarchy = next((item[2] for item in window if item[2]), [])
        section_type = window[0][3] if window else "section"
        start = cursor
        end = max(window_end, start)
        if end > start:
            ranges.append((start, end, hierarchy, section_type))
            cursor = end

    if cursor < total_len and ranges:
        start, end, hierarchy, section_type = ranges[-1]
        ranges[-1] = (start, total_len, hierarchy, section_type)
    elif cursor < total_len:
        ranges.append((0, total_len, [], "paragraph"))

    return ranges


def _filter_empty_ranges(
    content: str,
    ranges: list[tuple[int, int, list[str], str]],
) -> list[tuple[int, int, list[str], str]]:
    out: list[tuple[int, int, list[str], str]] = []
    for start, end, hierarchy, section_type in ranges:
        if end <= start:
            continue
        if not content[start:end].strip():
            if out:
                ps, pe, ph, pt = out[-1]
                out[-1] = (ps, end, ph, pt)
            continue
        out.append((start, end, hierarchy, section_type))
    return out


def _split_fenced_code_block(block_text: str) -> tuple[str, str, str, str] | None:
    lines = block_text.splitlines(keepends=True)
    if len(lines) < 2:
        return None
    opening = lines[0]
    if not opening.strip().startswith("```"):
        return None

    close_idx: int | None = None
    for idx in range(len(lines) - 1, 0, -1):
        if lines[idx].strip().startswith("```"):
            close_idx = idx
            break

    if close_idx is None:
        return None

    prefix = opening
    code = "".join(lines[1:close_idx])
    suffix = "".join(lines[close_idx:])

    fence_language = opening.strip()[3:].strip().lower()
    if code and not code.strip():
        prefix = prefix + code
        code = ""

    return prefix, code, suffix, fence_language


def _resolve_code_fence_language(fence_language: str, code_text: str) -> str | None:
    tag = fence_language.strip().lower()
    if tag:
        mapped = _CODE_FENCE_LANGUAGE_ALIASES.get(tag, tag)
        if mapped in _CODE_LANGUAGES:
            return mapped

    detected = detect_language(filepath="", content=code_text)
    if detected in _CODE_LANGUAGES:
        return detected

    return None


def _resolve_markup_fence_language(fence_language: str, block_text: str) -> str | None:
    tag = fence_language.strip().lower()
    if tag:
        mapped = _MARKUP_FENCE_LANGUAGE_ALIASES.get(tag, tag)
        if mapped in _MARKUP_LANGUAGES:
            return mapped

    detected = detect_language(filepath="", content=block_text)
    if detected in _MARKUP_LANGUAGES:
        return detected

    return None


def _rebase_delegated_fence_chunks(
    *,
    delegated_chunks: list[Chunk],
    heading_hierarchy: list[str],
    options: ChunkOptions,
    text_index: TextIndex,
    byte_offset: int,
    cumsum: Any,
    start_index: int,
) -> list[Chunk]:
    out: list[Chunk] = []
    for chunk in delegated_chunks:
        byte_start = byte_offset + chunk.byte_range.start
        byte_end = byte_offset + chunk.byte_range.end

        context = replace(
            chunk.context,
            heading_hierarchy=list(heading_hierarchy),
            section_type=EntityType.CODE_BLOCK.value,
        )
        contextualized_text = (
            chunk.text
            if options.context_mode == "none"
            else format_contextualized_text(chunk.text, context=context)
        )

        out.append(
            Chunk(
                text=chunk.text,
                contextualized_text=contextualized_text,
                byte_range=ByteRange(byte_start, byte_end),
                line_range=LineRange(
                    text_index.line_for_byte(byte_start),
                    text_index.line_for_byte(max(byte_start, byte_end - 1)),
                ),
                index=start_index + len(out),
                total_chunks=-1,
                context=context,
                token_count=chunk.token_count,
                char_count=len(chunk.text),
                nws_count=get_nws_count(cumsum, byte_start, byte_end),
            )
        )

    return out


def _build_markdown_fence_chunk(
    *,
    filepath: str,
    language: Language,
    text: str,
    heading_hierarchy: list[str],
    start_char: int,
    end_char: int,
    index: int,
    options: ChunkOptions,
    text_index: TextIndex,
    cumsum: Any,
    token_counter: Callable[[str], int],
) -> Chunk:
    byte_start = text_index.byte_offset_for_char(start_char)
    byte_end = text_index.byte_offset_for_char(end_char)
    line_start = text_index.line_for_char(start_char)
    line_end = text_index.line_for_char(max(start_char, end_char - 1))

    context = ChunkContext(
        filepath=filepath,
        language=language,
        content_type=ContentType.PROSE,
        heading_hierarchy=list(heading_hierarchy),
        section_type=EntityType.CODE_BLOCK.value,
        entities=_entities_for_section(
            EntityType.CODE_BLOCK.value,
            text,
            byte_start,
            byte_end,
            line_start,
            line_end,
        ),
    )

    contextualized_text = (
        text
        if options.context_mode == "none"
        else format_contextualized_text(text, context=context)
    )

    return Chunk(
        text=text,
        contextualized_text=contextualized_text,
        byte_range=ByteRange(byte_start, byte_end),
        line_range=LineRange(line_start, max(line_start, line_end)),
        index=index,
        total_chunks=-1,
        context=context,
        token_count=token_counter(text),
        char_count=len(text),
        nws_count=get_nws_count(cumsum, byte_start, byte_end),
    )


def _entities_for_section(
    section_type: str,
    text: str,
    start: int,
    end: int,
    line_start: int,
    line_end: int,
) -> list[EntityInfo]:
    mapping = {
        "heading": EntityType.HEADING,
        "section": EntityType.SECTION,
        "paragraph": EntityType.PARAGRAPH,
        "list": EntityType.LIST,
        "table": EntityType.TABLE,
        "code_block": EntityType.CODE_BLOCK,
        "frontmatter": EntityType.FRONTMATTER,
    }
    etype = mapping.get(section_type, EntityType.SECTION)
    signature = text.strip().splitlines()[0][:120] if text.strip() else section_type
    return [
        EntityInfo(
            name=section_type,
            type=etype,
            signature=signature,
            byte_range=ByteRange(start, end),
            line_range=LineRange(line_start, max(line_start, line_end)),
        )
    ]


def _finalize_chunk_indexes(chunks: list[Chunk]) -> list[Chunk]:
    total = len(chunks)
    return [
        Chunk(
            text=chunk.text,
            contextualized_text=chunk.contextualized_text,
            byte_range=chunk.byte_range,
            line_range=chunk.line_range,
            index=idx,
            total_chunks=total,
            context=chunk.context,
            token_count=chunk.token_count,
            char_count=chunk.char_count,
            nws_count=chunk.nws_count,
        )
        for idx, chunk in enumerate(chunks)
    ]


def _with_unknown_total(chunk: Chunk, index: int) -> Chunk:
    return Chunk(
        text=chunk.text,
        contextualized_text=chunk.contextualized_text,
        byte_range=chunk.byte_range,
        line_range=chunk.line_range,
        index=index,
        total_chunks=-1,
        context=chunk.context,
        token_count=chunk.token_count,
        char_count=chunk.char_count,
        nws_count=chunk.nws_count,
    )



from __future__ import annotations

import json
from collections.abc import Iterator
from pathlib import Path

from omnichunk import Chunker, chunk, chunk_directory, chunk_file
from omnichunk.engine.code_engine import CodeEngine
from omnichunk.types import Chunk, ChunkOptions


def test_chunk_function_simple_python() -> None:
    code = "def foo(x):\n    return x + 1\n"
    chunks = chunk("example.py", code, max_chunk_size=32, size_unit="chars")

    assert chunks
    assert all(c.text for c in chunks)
    assert chunks[0].context.filepath == "example.py"


def test_stream_is_lazy(monkeypatch: object) -> None:
    code = "def f():\n    return 1\n" * 8
    chunker = Chunker(max_chunk_size=48, min_chunk_size=8, size_unit="chars")
    yielded_before_second = {"n": 0}

    orig = CodeEngine._iter_base_chunks

    def _counting(
        self: CodeEngine,
        filepath: str,
        content: str,
        options: ChunkOptions,
    ) -> Iterator[Chunk]:
        for i, item in enumerate(orig(self, filepath, content, options)):
            yielded_before_second["n"] = i + 1
            yield item

    monkeypatch.setattr(CodeEngine, "_iter_base_chunks", _counting)

    iterator = chunker.stream("lazy.py", code)
    _ = next(iterator)
    assert yielded_before_second["n"] == 1
    _ = next(iterator)
    assert yielded_before_second["n"] == 2


def test_stream_total_chunks_is_minus_one() -> None:
    chunker = Chunker(max_chunk_size=40, size_unit="chars")
    text = "def g():\n    return 2\n"
    for c in chunker.stream("g.py", text):
        assert c.total_chunks == -1


def test_stream_reconstruction() -> None:
    code = "def foo():\n    return 1\n" * 15
    chunker = Chunker(max_chunk_size=64, min_chunk_size=10, size_unit="chars")
    out = "".join(c.text for c in chunker.stream("r.py", code))
    assert out == code


def test_chunker_stream_sets_unknown_total() -> None:
    text = "# Title\n\nHello world\n\nSecond paragraph."
    chunker = Chunker(max_chunk_size=20, size_unit="chars")

    streamed = list(chunker.stream("doc.md", text))
    assert streamed
    assert all(c.total_chunks == -1 for c in streamed)


def test_stream_text_matches_chunk_output_across_engines() -> None:
    chunker = Chunker(max_chunk_size=40, min_chunk_size=8, size_unit="chars")
    cases = [
        ("sample.py", "def add(a, b):\n    return a + b\n"),
        ("doc.md", "# Title\n\nParagraph one.\n\nParagraph two."),
        ("data.json", '{"a": 1, "b": {"c": 2}}'),
        (
            "cells.py",
            '# %% [markdown]\n"""\n# Intro\n"""\n# %%\nprint(\'hi\')\n',
        ),
    ]

    for filepath, content in cases:
        chunked = chunker.chunk(filepath, content)
        streamed = list(chunker.stream(filepath, content))

        assert [c.text for c in streamed] == [c.text for c in chunked]
        assert all(c.total_chunks == -1 for c in streamed)


def test_chunk_file_reads_disk(tmp_path: Path) -> None:
    file_path = tmp_path / "sample.py"
    file_path.write_text("def ping():\n    return 'pong'\n", encoding="utf-8")

    chunks = chunk_file(str(file_path), max_chunk_size=40, size_unit="chars")
    assert chunks
    assert "ping" in "\n".join(c.text for c in chunks)


def test_batch_keeps_input_order() -> None:
    chunker = Chunker(max_chunk_size=30, size_unit="chars")
    files = [
        {"filepath": "a.py", "code": "def a():\n    return 1\n"},
        {"filepath": "b.md", "code": "# B\n\nText"},
        {"filepath": "c.json", "code": '{"k": 1, "v": 2}'},
    ]

    results = chunker.batch(files, concurrency=2)

    assert [r.filepath for r in results] == ["a.py", "b.md", "c.json"]
    assert all(r.error is None for r in results)
    assert all(r.chunks for r in results)


def test_chunk_directory_glob_and_hidden_filters(tmp_path: Path) -> None:
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    (src_dir / "a.py").write_text("def a():\n    return 1\n", encoding="utf-8")
    (src_dir / "b.md").write_text("# title\n", encoding="utf-8")

    nested = src_dir / "nested"
    nested.mkdir()
    (nested / "c.py").write_text("def c():\n    return 3\n", encoding="utf-8")
    (nested / ".hidden.py").write_text("def hidden():\n    return 0\n", encoding="utf-8")

    results = chunk_directory(
        str(src_dir),
        glob="**/*.py",
        max_chunk_size=40,
        size_unit="chars",
    )

    filepaths = [Path(item.filepath).name for item in results]
    assert filepaths == ["a.py", "c.py"]
    assert all(item.error is None for item in results)
    assert all(item.chunks for item in results)


def test_chunker_export_and_quality_helpers() -> None:
    code = "def add(a: int, b: int) -> int:\n    return a + b\n"
    chunker = Chunker(max_chunk_size=48, min_chunk_size=12, size_unit="chars")
    chunks = chunker.chunk("calc.py", code)

    payload = chunker.to_dicts(chunks)
    assert payload
    assert payload[0]["context"]["filepath"] == "calc.py"

    jsonl = chunker.to_jsonl(chunks)
    lines = [line for line in jsonl.splitlines() if line.strip()]
    assert lines
    decoded = [json.loads(line) for line in lines]
    assert decoded[0]["context"]["filepath"] == "calc.py"

    csv_payload = chunker.to_csv(chunks)
    assert "filepath" in csv_payload.splitlines()[0]
    assert "calc.py" in csv_payload

    stats = chunker.chunk_stats(chunks, size_unit="chars")
    assert stats.total_chunks == len(chunks)
    assert stats.average_size > 0

    quality = chunker.quality_scores(
        chunks,
        min_chunk_size=12,
        max_chunk_size=48,
        size_unit="chars",
    )
    assert quality
    assert all(0.0 <= item.overall <= 1.0 for item in quality)


def test_chunker_with_python_nws_backend_keeps_integrity() -> None:
    content = (
        "def a(x: int) -> int:\n"
        "    return x + 1\n\n\n"
        "def b(y: int) -> int:\n"
        "    return y + 2\n"
    )
    chunker = Chunker(
        max_chunk_size=42,
        min_chunk_size=10,
        size_unit="chars",
        nws_backend="python",
    )

    chunks = chunker.chunk("module.py", content)
    assert chunks
    assert "".join(chunk.text for chunk in chunks) == content
    for left, right in zip(chunks, chunks[1:]):
        assert left.byte_range.end == right.byte_range.start



from __future__ import annotations

import importlib
import io
from typing import Any

from omnichunk.formats.types import FormatSegment, LoadedDocument


def load_docx_bytes(raw: bytes) -> LoadedDocument:
    """Extract text from DOCX bytes into prose segments (paragraphs and tables)."""
    try:
        Document = importlib.import_module("docx").Document
        CT_Tbl = importlib.import_module("docx.oxml.table").CT_Tbl
        CT_P = importlib.import_module("docx.oxml.text.paragraph").CT_P
        Table = importlib.import_module("docx.table").Table
        Paragraph = importlib.import_module("docx.text.paragraph").Paragraph
    except ImportError as exc:
        raise ImportError(
            "DOCX support requires python-docx. Install with: pip install omnichunk[docx]"
        ) from exc

    warnings: list[str] = []
    document = Document(io.BytesIO(raw))

    def iter_block_items(parent: Any) -> Any:
        parent_elm = parent.element.body
        for child in parent_elm.iterchildren():
            if isinstance(child, CT_P):
                yield Paragraph(child, parent)
            elif isinstance(child, CT_Tbl):
                yield Table(child, parent)

    parts: list[str] = []
    segments: list[FormatSegment] = []
    cursor = 0
    block_index = 0

    for block in iter_block_items(document):
        if isinstance(block, Paragraph):
            ptext = block.text
            if not ptext.strip():
                block_index += 1
                continue
            style_name = ""
            try:
                style_name = block.style.name if block.style is not None else ""
            except Exception:
                warnings.append(f"paragraph_style_{block_index}")
            start = cursor
            chunk = ptext + "\n\n"
            parts.append(chunk)
            cursor += len(chunk)
            meta: dict[str, Any] = {"docx_block": "paragraph", "index": block_index}
            if style_name and "Heading" in style_name:
                meta["heading_style"] = style_name
            segments.append(
                FormatSegment(
                    char_start=start,
                    char_end=cursor,
                    kind="prose",
                    metadata=meta,
                )
            )
        else:
            table = block
            rows_text: list[str] = []
            for row in table.rows:
                cells = [c.text.strip() for c in row.cells]
                rows_text.append("\t".join(cells))
            tbl = "\n".join(rows_text) + "\n\n"
            if not tbl.strip():
                block_index += 1
                continue
            start = cursor
            parts.append(tbl)
            cursor += len(tbl)
            segments.append(
                FormatSegment(
                    char_start=start,
                    char_end=cursor,
                    kind="prose",
                    metadata={"docx_block": "table", "index": block_index},
                )
            )
        block_index += 1

    full_text = "".join(parts)
    return LoadedDocument(
        text=full_text,
        segments=tuple(segments),
        format_name="docx",
        warnings=tuple(warnings),
    )



from __future__ import annotations

from pathlib import Path

from omnichunk import Chunker
from omnichunk.context import entities as entities_mod
from omnichunk.context.siblings import build_sibling_index, detect_siblings_for_chunk
from omnichunk.types import ByteRange, EntityInfo, EntityType, LineRange


class _FakeNode:
    def __init__(
        self,
        node_type: str,
        start: int,
        end: int,
        *,
        children: list[_FakeNode] | None = None,
        fields: dict[str, _FakeNode] | None = None,
        is_named: bool = True,
    ) -> None:
        self.type = node_type
        self.start_byte = start
        self.end_byte = end
        self.start_point = (0, 0)
        self.end_point = (0, 0)
        self.children = children or []
        self.named_children = [
            child for child in self.children if getattr(child, "is_named", True)
        ]
        self._fields = fields or {}
        self.is_named = is_named
        self.parent: _FakeNode | None = None

        for child in self.children:
            child.parent = self
        for child in self._fields.values():
            child.parent = self

    def child_by_field_name(self, field_name: str) -> _FakeNode | None:
        return self._fields.get(field_name)


def test_contextualized_text_contains_metadata(fixtures_dir: Path) -> None:
    code = (fixtures_dir / "python_complex.py").read_text(encoding="utf-8")
    chunker = Chunker(max_chunk_size=240, size_unit="chars", context_mode="full")

    chunks = chunker.chunk("src/services/user_service.py", code)

    assert chunks
    sample = chunks[0].contextualized_text
    assert "# src/services/user_service.py" in sample
    assert "# Language: python" in sample


def test_import_tracking_and_siblings() -> None:
    code = (
        "import os\n"
        "import json\n\n"
        "def a():\n    return os.getcwd()\n\n"
        "def b():\n    return json.dumps({})\n\n"
        "def c():\n    return 3\n"
    )
    chunker = Chunker(max_chunk_size=55, min_chunk_size=10, size_unit="chars")

    chunks = chunker.chunk("imports.py", code)

    assert chunks
    assert any(c.context.imports for c in chunks)
    assert any(c.context.siblings for c in chunks)


def test_filter_imports_option() -> None:
    code = "import os\nimport json\n\ndef a():\n    return os.getcwd()\n"
    chunker = Chunker(max_chunk_size=80, size_unit="chars", filter_imports=True)

    chunks = chunker.chunk("imports.py", code)

    assert chunks
    for chunk in chunks:
        if chunk.context.imports:
            names = {imp.name for imp in chunk.context.imports}
            assert names <= {"os", "json"}


def test_siblings_prefer_same_parent_scope() -> None:
    code = (
        "class Service:\n"
        "    def first(self):\n"
        "        return 1\n\n"
        "    def second(self):\n"
        "        return 2\n\n"
        "def standalone():\n"
        "    return 3\n"
    )
    chunker = Chunker(max_chunk_size=35, min_chunk_size=5, size_unit="chars", max_siblings=3)

    chunks = chunker.chunk("scope_siblings.py", code)

    target = next((c for c in chunks if "def second" in c.text), None)
    assert target is not None

    names = {s.name for s in target.context.siblings}
    assert "first" in names


def test_python_entity_signatures_include_full_type_hints() -> None:
    code = (
        "from typing import Optional\n\n"
        "class Service:\n"
        "    def process(self, data: dict, limit: int = 10) -> Optional[str]:\n"
        "        return None\n"
    )
    chunker = Chunker(max_chunk_size=300, size_unit="chars")
    chunks = chunker.chunk("service.py", code)
    all_entities = [e for c in chunks for e in c.context.entities]

    process_entity = next((e for e in all_entities if e.name == "process"), None)
    assert process_entity is not None
    assert "data: dict" in process_entity.signature
    assert "limit: int" in process_entity.signature


def test_python_class_signature_with_bases() -> None:
    code = "class MyService(BaseService, Mixin):\n    pass\n"
    chunker = Chunker(max_chunk_size=200, size_unit="chars")
    chunks = chunker.chunk("svc.py", code)
    all_entities = [e for c in chunks for e in c.context.entities]

    cls = next((e for e in all_entities if e.name == "MyService"), None)
    assert cls is not None
    assert "BaseService" in cls.signature
    assert "Mixin" in cls.signature


def test_python_docstring_requires_first_statement_string() -> None:
    good_source = '"""Actual docstring."""\nreturn 1\n'
    good_literal = _FakeNode("string", 0, len('"""Actual docstring."""'))
    good_stmt = _FakeNode(
        "expression_statement",
        0,
        len('"""Actual docstring."""'),
        children=[good_literal],
    )
    return_stmt = _FakeNode("return_statement", len('"""Actual docstring."""\n'), len(good_source))
    good_body = _FakeNode("block", 0, len(good_source), children=[good_stmt, return_stmt])
    good_func = _FakeNode(
        "function_definition",
        0,
        len(good_source),
        fields={"body": good_body},
    )

    bad_source = 'value = """not a docstring"""\n"""also not a docstring"""\n'
    assignment = _FakeNode("assignment", 0, bad_source.index("\n") + 1)
    trailing_literal = _FakeNode("string", bad_source.index('"""also'), len(bad_source) - 1)
    trailing_stmt = _FakeNode(
        "expression_statement",
        bad_source.index('"""also'),
        len(bad_source),
        children=[trailing_literal],
    )
    bad_body = _FakeNode("block", 0, len(bad_source), children=[assignment, trailing_stmt])
    bad_func = _FakeNode("function_definition", 0, len(bad_source), fields={"body": bad_body})

    assert entities_mod._extract_python_docstring(good_func, good_source.encode("utf-8")) == (
        "Actual docstring."
    )
    assert entities_mod._extract_python_docstring(bad_func, bad_source.encode("utf-8")) is None


def test_javascript_jsdoc_attaches_only_to_adjacent_entity() -> None:
    source = (
        "/** Detached comment */\n"
        "const value = 1;\n\n"
        "/**\n"
        " * Attached to run\n"
        " */\n"
        "function run() {}\n"
        "function noop() {}\n"
    )

    detached_start = source.index("/** Detached comment */")
    detached_end = detached_start + len("/** Detached comment */")
    const_start = source.index("const value")
    const_end = source.index(";", const_start) + 1
    attached_start = source.index("/**\n")
    attached_end = source.index("*/", attached_start) + 2
    run_start = source.index("function run")
    run_end = source.index("}\n", run_start) + 1
    noop_start = source.index("function noop")
    noop_end = source.rindex("}") + 1

    detached_comment = _FakeNode("comment", detached_start, detached_end)
    const_decl = _FakeNode("lexical_declaration", const_start, const_end)
    attached_comment = _FakeNode("comment", attached_start, attached_end)
    run_func = _FakeNode("function_declaration", run_start, run_end)
    noop_func = _FakeNode("function_declaration", noop_start, noop_end)

    _FakeNode(
        "program",
        0,
        len(source),
        children=[detached_comment, const_decl, attached_comment, run_func, noop_func],
    )

    doc = entities_mod._extract_leading_comment(run_func, source.encode("utf-8"))
    assert doc == "Attached to run"
    assert entities_mod._extract_leading_comment(noop_func, source.encode("utf-8")) is None


def test_rust_grouped_imports_expand_to_individual_items() -> None:
    imports_one = entities_mod._parse_rust_imports("use std::collections::{HashMap, HashSet};")
    imports_two = entities_mod._parse_rust_imports(
        "use crate::net::{self, client::Client as HttpClient};"
    )
    import_names = {name for name, _ in [*imports_one, *imports_two]}

    assert {"HashMap", "HashSet", "net", "HttpClient"} <= import_names


def test_go_import_block_parses_each_import_spec() -> None:
    code = (
        "package main\n\n"
        "import (\n"
        '    "fmt"\n'
        '    osalias "os"\n'
        ")\n\n"
        "func main() {\n"
        "    fmt.Println(osalias.Args)\n"
        "}\n"
    )

    parsed = entities_mod._parse_go_imports(code)
    import_names = {name for name, _ in parsed}

    assert {"fmt", "osalias"} <= import_names


def test_java_imports_include_static_and_non_static_symbols() -> None:
    code = (
        "import java.util.List;\n"
        "import static java.lang.Math.max;\n\n"
        "public class Main {\n"
        "    int clamp(int v) {\n"
        "        return max(v, 0);\n"
        "    }\n"
        "}\n"
    )

    parsed = entities_mod._parse_java_imports(code)
    import_names = {name for name, _ in parsed}

    assert {"List", "max"} <= import_names


def test_sibling_index_matches_list_detection() -> None:
    entities = [
        EntityInfo(
            name="A",
            type=EntityType.CLASS,
            byte_range=ByteRange(0, 200),
            line_range=LineRange(0, 20),
        ),
        EntityInfo(
            name="first",
            type=EntityType.METHOD,
            parent="A",
            byte_range=ByteRange(10, 40),
            line_range=LineRange(1, 4),
        ),
        EntityInfo(
            name="second",
            type=EntityType.METHOD,
            parent="A",
            byte_range=ByteRange(45, 80),
            line_range=LineRange(5, 8),
        ),
        EntityInfo(
            name="third",
            type=EntityType.METHOD,
            parent="A",
            byte_range=ByteRange(90, 130),
            line_range=LineRange(9, 12),
        ),
    ]
    chunk_range = ByteRange(46, 79)

    from_list = detect_siblings_for_chunk(entities, chunk_range, max_siblings=2)
    sibling_index = build_sibling_index(entities)
    from_index = detect_siblings_for_chunk(sibling_index, chunk_range, max_siblings=2)

    assert from_index == from_list
    assert {item.name for item in from_index} == {"first", "third"}


def test_find_block_end_respects_next_top_level_definition() -> None:
    code = (
        "def alpha():\n"
        "    if True:\n"
        "        return 1\n"
        "\n"
        "    return 2\n"
        "\n"
        "def beta():\n"
        "    return 3\n"
    )

    start = code.index("def alpha")
    end = entities_mod._find_block_end(code, start)

    assert code[end:].startswith("def beta")



from __future__ import annotations

import sys
from dataclasses import dataclass
from pathlib import Path

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

from run_benchmarks import SCENARIOS, Scenario  # noqa: E402

from omnichunk import Chunk, Chunker  # noqa: E402


@dataclass(frozen=True)
class QualityRow:
    scenario: str
    chunk_count: int
    reconstruction_ok: bool
    contiguous_ok: bool
    non_empty_ok: bool
    deterministic_ok: bool


def _is_contiguous_non_overlapping(chunks: list[Chunk]) -> bool:
    if not chunks:
        return True

    ordered = sorted(chunks, key=lambda c: c.byte_range.start)
    cursor = ordered[0].byte_range.start
    for chunk in ordered:
        start = chunk.byte_range.start
        end = chunk.byte_range.end
        if end < start:
            return False
        if start != cursor:
            return False
        cursor = end
    return True


def _is_deterministic(left: list[Chunk], right: list[Chunk]) -> bool:
    if len(left) != len(right):
        return False

    for lchunk, rchunk in zip(left, right):
        if lchunk.text != rchunk.text:
            return False
        if lchunk.contextualized_text != rchunk.contextualized_text:
            return False
        if lchunk.byte_range != rchunk.byte_range:
            return False
        if lchunk.line_range != rchunk.line_range:
            return False
        if lchunk.context != rchunk.context:
            return False

    return True


def evaluate_scenario(chunker: Chunker, scenario: Scenario) -> QualityRow:
    text = scenario.path.read_text(encoding="utf-8")
    chunks = chunker.chunk(
        str(scenario.path),
        text,
        max_chunk_size=scenario.max_chunk_size,
        min_chunk_size=scenario.min_chunk_size,
        size_unit=scenario.size_unit,
    )
    repeated = chunker.chunk(
        str(scenario.path),
        text,
        max_chunk_size=scenario.max_chunk_size,
        min_chunk_size=scenario.min_chunk_size,
        size_unit=scenario.size_unit,
    )

    reconstruction_ok = "".join(c.text for c in chunks) == text
    contiguous_ok = _is_contiguous_non_overlapping(chunks)
    non_empty_ok = all(c.text.strip() for c in chunks)
    deterministic_ok = _is_deterministic(chunks, repeated)

    return QualityRow(
        scenario=scenario.name,
        chunk_count=len(chunks),
        reconstruction_ok=reconstruction_ok,
        contiguous_ok=contiguous_ok,
        non_empty_ok=non_empty_ok,
        deterministic_ok=deterministic_ok,
    )


def run() -> int:
    chunker = Chunker()
    print("Scenario,Chunks,Reconstruction,Contiguous,NonEmpty,Deterministic")

    rows: list[QualityRow] = [evaluate_scenario(chunker, scenario) for scenario in SCENARIOS]
    for row in rows:
        print(
            f"{row.scenario},{row.chunk_count},"
            f"{str(row.reconstruction_ok).lower()},"
            f"{str(row.contiguous_ok).lower()},"
            f"{str(row.non_empty_ok).lower()},"
            f"{str(row.deterministic_ok).lower()}"
        )

    all_ok = all(
        row.reconstruction_ok and row.contiguous_ok and row.non_empty_ok and row.deterministic_ok
        for row in rows
    )
    print(f"TOTAL,{len(rows)},status={'ok' if all_ok else 'fail'}")
    return 0 if all_ok else 1


if __name__ == "__main__":
    raise SystemExit(run())



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

import numpy as np

from omnichunk.sizing.counter import make_size_counter, make_token_counter
from omnichunk.sizing.nws import get_nws_count
from omnichunk.types import (
    ByteRange,
    Chunk,
    ChunkContext,
    ChunkOptions,
    ContentType,
    LineRange,
)
from omnichunk.util.detect import detect_language
from omnichunk.util.text_index import TextIndex

from .boundaries import detect_semantic_boundaries
from .sentences import split_sentences


@dataclass(frozen=True)
class SemanticSplitter:
    """Splits prose text into semantically coherent chunks.

    embed_fn must accept list[str] and return np.ndarray of shape (N, D).
    window: number of sentences per context window for boundary detection.
    threshold: cosine similarity below which a boundary is placed.
    min_chunk_sentences: minimum sentences per output chunk.
    sentence_splitter_fn: optional custom sentence splitter.
    """

    embed_fn: Callable[[list[str]], np.ndarray]
    window: int = 3
    threshold: float = 0.3
    min_chunk_sentences: int = 1
    sentence_splitter_fn: Callable[[str], list[str]] | None = None

    def split(
        self,
        filepath: str,
        text: str,
        options: ChunkOptions,
    ) -> list[Chunk]:
        if not text.strip():
            return []

        text_index = options._precomputed_text_index
        if text_index is None:
            text_index = TextIndex(text)
        cumsum = options._precomputed_nws_cumsum
        if cumsum is None:
            raise RuntimeError("SemanticSplitter requires precomputed NWS cumsum on options")

        language = options.language or detect_language(filepath=filepath, content=text)

        triples = split_sentences(text, splitter_fn=self.sentence_splitter_fn)
        sentence_texts = [t for t, _, _ in triples]
        if not sentence_texts:
            return []

        boundary_res = detect_semantic_boundaries(
            sentence_texts,
            embed_fn=self.embed_fn,
            window=self.window,
            threshold=self.threshold,
            min_chunk_sentences=max(1, self.min_chunk_sentences),
        )
        boundaries = list(boundary_res.boundary_indices)
        n = len(triples)

        starts = [0]
        for b in boundaries:
            starts.append(b + 1)
        ends = list(boundaries) + [n - 1]

        groups: list[tuple[int, int]] = []
        if len(starts) != len(ends):
            raise RuntimeError("semantic splitter: starts/ends length mismatch")
        for s, e in zip(starts, ends):
            if s <= e:
                groups.append((s, e))

        size_counter = make_size_counter(
            options.size_unit,
            options.tokenizer,
            max_token_chars=256,
            chunk_size=options.max_chunk_size,
        )
        token_counter = make_token_counter(
            options.tokenizer,
            max_token_chars=256,
            chunk_size=options.max_chunk_size,
        )

        def piece_size(sub: str) -> int:
            return int(size_counter(sub))

        def subdivide_sentence_indices(s: int, e: int) -> list[tuple[int, int]]:
            stack = [(s, e)]
            final: list[tuple[int, int]] = []
            while stack:
                a, b = stack.pop()
                chunk_str = "".join(triples[i][0] for i in range(a, b + 1))
                if piece_size(chunk_str) <= options.max_chunk_size:
                    final.append((a, b))
                    continue
                if a < b:
                    mid = (a + b + 1) // 2
                    stack.append((mid, b))
                    stack.append((a, mid - 1))
                    continue
                final.append((a, b))
            final.sort(key=lambda x: (x[0], x[1]))
            return final

        def char_piece_ranges(c0: int, c1: int) -> list[tuple[int, int]]:
            """Split [c0, c1) char range until each piece fits max_chunk_size."""
            stack = [(c0, c1)]
            parts: list[tuple[int, int]] = []
            while stack:
                a, b = stack.pop()
                if a >= b:
                    continue
                sub = text[a:b]
                if piece_size(sub) <= options.max_chunk_size:
                    parts.append((a, b))
                    continue
                if b - a <= 1:
                    parts.append((a, b))
                    continue
                mid = (a + b) // 2
                stack.append((mid, b))
                stack.append((a, mid))
            parts.sort(key=lambda x: (x[0], x[1]))
            return parts

        expanded_chars: list[tuple[int, int]] = []
        for s, e in groups:
            for a, b in subdivide_sentence_indices(s, e):
                chunk_str = "".join(triples[i][0] for i in range(a, b + 1))
                c0, c1 = triples[a][1], triples[b][2]
                if piece_size(chunk_str) <= options.max_chunk_size:
                    expanded_chars.append((c0, c1))
                else:
                    expanded_chars.extend(char_piece_ranges(c0, c1))

        chunks: list[Chunk] = []
        for idx, (c_start, c_end) in enumerate(expanded_chars):
            chunk_text = text[c_start:c_end]
            if not chunk_text.strip():
                continue
            byte_start = text_index.byte_offset_for_char(c_start)
            byte_end = text_index.byte_offset_for_char(c_end)
            line_start = text_index.line_for_char(c_start)
            line_end = text_index.line_for_char(max(c_start, c_end - 1))

            context = ChunkContext(
                filepath=filepath,
                language=language,
                content_type=ContentType.PROSE,
            )
            chunks.append(
                Chunk(
                    text=chunk_text,
                    contextualized_text=chunk_text,
                    byte_range=ByteRange(byte_start, byte_end),
                    line_range=LineRange(line_start, max(line_start, line_end)),
                    index=idx,
                    total_chunks=-1,
                    context=context,
                    token_count=int(token_counter(chunk_text)),
                    char_count=len(chunk_text),
                    nws_count=get_nws_count(cumsum, byte_start, byte_end),
                )
            )

        total = len(chunks)
        return [
            Chunk(
                text=c.text,
                contextualized_text=c.contextualized_text,
                byte_range=c.byte_range,
                line_range=c.line_range,
                index=i,
                total_chunks=total,
                context=c.context,
                token_count=c.token_count,
                char_count=c.char_count,
                nws_count=c.nws_count,
            )
            for i, c in enumerate(chunks)
        ]



from __future__ import annotations

import time
from collections.abc import Iterator
from contextlib import contextmanager, suppress
from typing import Any


def _span_set(span: object | None, key: str, value: Any) -> None:
    if span is None:
        return
    setter = getattr(span, "set_attribute", None)
    if setter is None:
        return
    with suppress(Exception):
        setter(key, value)


@contextmanager
def maybe_span(tracer: object | None, name: str, **initial_attrs: Any) -> Iterator[Any]:
    """OpenTelemetry span context manager; no-op when ``tracer`` is None."""
    if tracer is None:
        yield None
        return
    enter = getattr(tracer, "start_as_current_span", None)
    if enter is None:
        yield None
        return
    with enter(name) as span:
        for k, v in initial_attrs.items():
            _span_set(span, k, v)
        yield span


def finalize_chunk_file_span(
    span: object | None,
    *,
    chunk_count: int,
    t0: float,
    error: str | None = None,
) -> None:
    """Set duration and counts on an active chunk_file span."""
    dur = time.perf_counter() - t0
    _span_set(span, "omnichunk.chunking_duration_ms", round(dur * 1000.0, 3))
    _span_set(span, "omnichunk.chunk_count", chunk_count)
    if error:
        _span_set(span, "omnichunk.parse_errors", error[:2000])



from __future__ import annotations

from bisect import bisect_left
from dataclasses import dataclass

from omnichunk.types import ByteRange, EntityInfo, SiblingInfo

_IGNORED_ENTITY_TYPES = {"import", "export"}


@dataclass(frozen=True)
class _SiblingRow:
    entity: EntityInfo
    start: int
    end: int


@dataclass(frozen=True)
class _ParentGroup:
    rows: tuple[_SiblingRow, ...]
    global_to_local: dict[int, int]


@dataclass(frozen=True)
class SiblingIndex:
    rows: tuple[_SiblingRow, ...]
    starts: tuple[int, ...]
    prefix_max_ends: tuple[int, ...]
    parent_groups: dict[str, _ParentGroup]


def build_sibling_index(entities: list[EntityInfo]) -> SiblingIndex:
    rows: list[_SiblingRow] = []
    for entity in entities:
        byte_range = entity.byte_range
        if byte_range is None:
            continue
        if entity.type.value in _IGNORED_ENTITY_TYPES:
            continue
        rows.append(
            _SiblingRow(
                entity=entity,
                start=int(byte_range.start),
                end=int(byte_range.end),
            )
        )

    rows.sort(key=lambda row: (row.start, row.end, row.entity.name))
    sorted_rows = tuple(rows)
    starts = tuple(row.start for row in sorted_rows)
    prefix_max_ends = _build_prefix_max_ends(sorted_rows)
    parent_groups = _build_parent_groups(sorted_rows)

    return SiblingIndex(
        rows=sorted_rows,
        starts=starts,
        prefix_max_ends=prefix_max_ends,
        parent_groups=parent_groups,
    )


def detect_siblings_for_chunk(
    entities: list[EntityInfo] | SiblingIndex,
    chunk_range: ByteRange,
    *,
    max_siblings: int = 3,
) -> list[SiblingInfo]:
    """Detect neighboring entities around a chunk, ordered by distance."""
    if max_siblings <= 0:
        return []

    sibling_index = (
        entities if isinstance(entities, SiblingIndex) else build_sibling_index(entities)
    )
    if not sibling_index.rows:
        return []

    anchor_global_idx = _resolve_anchor_index(sibling_index, chunk_range)

    active_rows = sibling_index.rows
    anchor_local_idx = anchor_global_idx

    anchor_parent = active_rows[anchor_global_idx].entity.parent
    if anchor_parent:
        group = sibling_index.parent_groups.get(anchor_parent)
        if group is not None:
            active_rows = group.rows
            anchor_local_idx = group.global_to_local.get(anchor_global_idx, anchor_local_idx)

    before = _collect_directional_siblings(
        active_rows,
        chunk_range,
        anchor_local_idx,
        direction=-1,
        max_siblings=max_siblings,
    )
    after = _collect_directional_siblings(
        active_rows,
        chunk_range,
        anchor_local_idx,
        direction=1,
        max_siblings=max_siblings,
    )

    return before + after


def _build_prefix_max_ends(rows: tuple[_SiblingRow, ...]) -> tuple[int, ...]:
    prefix: list[int] = []
    running = -1
    for row in rows:
        if row.end > running:
            running = row.end
        prefix.append(running)
    return tuple(prefix)


def _build_parent_groups(rows: tuple[_SiblingRow, ...]) -> dict[str, _ParentGroup]:
    grouped: dict[str, list[tuple[int, _SiblingRow]]] = {}
    for global_idx, row in enumerate(rows):
        parent = row.entity.parent
        if not parent:
            continue
        grouped.setdefault(parent, []).append((global_idx, row))

    out: dict[str, _ParentGroup] = {}
    for parent, items in grouped.items():
        if len(items) < 2:
            continue
        parent_rows = tuple(row for _, row in items)
        global_to_local = {global_idx: local_idx for local_idx, (global_idx, _) in enumerate(items)}
        out[parent] = _ParentGroup(rows=parent_rows, global_to_local=global_to_local)
    return out


def _resolve_anchor_index(sibling_index: SiblingIndex, chunk_range: ByteRange) -> int:
    overlap_idx = _find_first_overlap_index(sibling_index, chunk_range)
    if overlap_idx >= 0:
        return overlap_idx
    return _nearest_index_by_start(sibling_index.starts, chunk_range.start)


def _find_first_overlap_index(sibling_index: SiblingIndex, chunk_range: ByteRange) -> int:
    right = bisect_left(sibling_index.starts, chunk_range.end)
    if right <= 0:
        return -1

    left = bisect_left(sibling_index.prefix_max_ends, chunk_range.start + 1, hi=right)
    for idx in range(left, right):
        row = sibling_index.rows[idx]
        if _ranges_overlap(row.start, row.end, chunk_range.start, chunk_range.end):
            return idx
    return -1


def _nearest_index_by_start(starts: tuple[int, ...], target_start: int) -> int:
    if not starts:
        return 0

    idx = bisect_left(starts, target_start)
    if idx <= 0:
        return 0
    if idx >= len(starts):
        return len(starts) - 1

    prev_idx = idx - 1
    prev_dist = abs(starts[prev_idx] - target_start)
    curr_dist = abs(starts[idx] - target_start)
    if curr_dist < prev_dist:
        return idx
    return prev_idx


def _collect_directional_siblings(
    rows: tuple[_SiblingRow, ...],
    chunk_range: ByteRange,
    anchor_idx: int,
    *,
    direction: int,
    max_siblings: int,
) -> list[SiblingInfo]:
    siblings: list[SiblingInfo] = []
    distance = 0

    idx = anchor_idx + direction
    while 0 <= idx < len(rows) and distance < max_siblings:
        row = rows[idx]
        if not _ranges_overlap(row.start, row.end, chunk_range.start, chunk_range.end):
            distance += 1
            siblings.append(
                SiblingInfo(
                    name=row.entity.name,
                    type=row.entity.type,
                    position="before" if direction < 0 else "after",
                    distance=distance,
                    signature=row.entity.signature,
                )
            )

        idx += direction

    return siblings


def _ranges_overlap(a_start: int, a_end: int, b_start: int, b_end: int) -> bool:
    return not (a_end <= b_start or a_start >= b_end)



from __future__ import annotations

import importlib
import os
from functools import lru_cache
from typing import Any, Literal, TypedDict

import numpy as np

NwsBackend = Literal["auto", "python", "rust"]

NWS_BACKEND_ENV_VAR = "OMNICHUNK_NWS_BACKEND"

_VALID_BACKENDS: dict[str, NwsBackend] = {
    "auto": "auto",
    "python": "python",
    "rust": "rust",
}


class NwsBackendStatus(TypedDict):
    selected: NwsBackend
    available: bool
    active: bool
    reason: str


def resolve_nws_backend(backend: str | None = None) -> NwsBackend:
    raw = backend if backend is not None else os.getenv(NWS_BACKEND_ENV_VAR, "auto")
    normalized = raw.strip().lower()
    resolved = _VALID_BACKENDS.get(normalized)
    if resolved is None:
        expected = ", ".join(sorted(_VALID_BACKENDS))
        raise ValueError(
            f"Invalid NWS backend '{raw}'. Expected one of: {expected}. "
            f"Use the {NWS_BACKEND_ENV_VAR} environment variable or per-call override."
        )
    return resolved


@lru_cache(maxsize=1)
def _load_rust_module() -> tuple[Any | None, str]:
    try:
        module = importlib.import_module("omnichunk_rust")
    except Exception as exc:
        return None, f"{type(exc).__name__}: {exc}"

    fn = getattr(module, "preprocess_nws_cumsum_bytes", None)
    if not callable(fn):
        return None, "omnichunk_rust.preprocess_nws_cumsum_bytes is unavailable"

    return module, ""


def reset_rust_backend_cache() -> None:
    _load_rust_module.cache_clear()


def nws_backend_status(backend: str | None = None) -> NwsBackendStatus:
    selected = resolve_nws_backend(backend)
    module, reason = _load_rust_module()
    available = module is not None
    active = selected == "rust" or (selected == "auto" and available)
    if selected == "python":
        active = False

    return {
        "selected": selected,
        "available": available,
        "active": active,
        "reason": "" if available else reason,
    }


def batch_cosine_similarity_adjacent_rust(
    embeddings: np.ndarray,
) -> np.ndarray | None:
    """Rust-accelerated adjacent cosine similarity when the extension is built.

    ``embeddings``: shape (N, D), float32 or float64. Returns shape (N-1,) float64,
    or None if Rust is unavailable.
    """
    module, _ = _load_rust_module()
    if module is None:
        return None
    fn = getattr(module, "batch_cosine_similarity_adjacent", None)
    if not callable(fn):
        return None
    arr = np.asarray(embeddings, dtype=np.float32, order="C")
    if arr.ndim != 2 or arr.shape[1] == 0:
        return None
    _n, d = arr.shape
    flat_list = arr.ravel().tolist()
    try:
        result = fn(flat_list, int(d))
    except Exception:
        return None
    return np.asarray(result, dtype=np.float64)


def maybe_preprocess_nws_cumsum_rust(
    code: str,
    *,
    backend: str | None = None,
) -> np.ndarray | None:
    selected = resolve_nws_backend(backend)
    if selected == "python":
        return None

    module, reason = _load_rust_module()
    if module is None:
        if selected == "rust":
            raise RuntimeError(
                "Rust NWS backend was explicitly requested but is unavailable: "
                f"{reason}. Build and install rust/omnichunk_rust first."
            )
        return None

    raw = code.encode("utf-8")
    values = module.preprocess_nws_cumsum_bytes(raw)
    return np.asarray(values, dtype=np.int64)



from __future__ import annotations

import pytest

from omnichunk import Chunker


def test_hierarchical_chunk_levels_validated() -> None:
    chunker = Chunker()
    with pytest.raises(ValueError):
        chunker.hierarchical_chunk("f.py", "x", levels=[64])
    with pytest.raises(ValueError):
        chunker.hierarchical_chunk("f.py", "x", levels=[256, 64])


def test_hierarchical_chunk_leaf_reconstruction() -> None:
    code = "def foo():\n    return 1\n" * 20
    chunker = Chunker()
    tree = chunker.hierarchical_chunk("m.py", code, levels=[32, 128, 512], size_unit="chars")
    leaves = tree.leaves()
    assert leaves
    assert "".join(c.text for c in leaves) == code
    raw = code.encode("utf-8")
    for c in leaves:
        assert raw[c.byte_range.start : c.byte_range.end].decode("utf-8") == c.text
    for left, right in zip(leaves, leaves[1:]):
        assert left.byte_range.end == right.byte_range.start


def test_hierarchical_chunk_parent_spans_children() -> None:
    code = "def f():\n    pass\n" * 30
    chunker = Chunker()
    tree = chunker.hierarchical_chunk("m.py", code, levels=[24, 96], size_unit="chars")
    for node in tree.nodes:
        if node.level > 0:
            children = tree.children(node.chunk)
            if children:
                assert node.chunk.byte_range.start == children[0].byte_range.start
                assert node.chunk.byte_range.end == children[-1].byte_range.end


def test_hierarchical_chunk_level_count() -> None:
    text = "Hello world. " * 50
    chunker = Chunker()
    tree = chunker.hierarchical_chunk("doc.md", text, levels=[16, 64, 256], size_unit="chars")
    assert tree.level_count == 3


def test_hierarchical_chunk_at_level_returns_sorted() -> None:
    text = "Para one.\n\n" * 40
    chunker = Chunker()
    tree = chunker.hierarchical_chunk("doc.md", text, levels=[20, 80], size_unit="chars")
    for level in range(tree.level_count):
        chunks = tree.at_level(level)
        for left, right in zip(chunks, chunks[1:]):
            assert left.byte_range.start < right.byte_range.start


def test_hierarchical_chunk_two_levels_tree_structure() -> None:
    text = "Sentence one. " * 60
    chunker = Chunker()
    tree = chunker.hierarchical_chunk("doc.md", text, levels=[20, 80], size_unit="chars")
    leaves = tree.leaves()
    roots = tree.roots()
    assert len(roots) <= len(leaves)
    for root in roots:
        children = tree.children(root)
        assert children


def test_hierarchical_chunk_to_dict_round_trip() -> None:
    code = "def a():\n    pass\n" * 10
    chunker = Chunker()
    tree = chunker.hierarchical_chunk("m.py", code, levels=[24, 96], size_unit="chars")
    d = tree.to_dict()
    assert d["level_count"] == 2
    assert len(d["nodes"]) == len(tree.nodes)


def test_hierarchical_chunk_deterministic() -> None:
    code = "def f(x):\n    return x\n" * 15
    chunker = Chunker()
    t1 = chunker.hierarchical_chunk("m.py", code, levels=[32, 128], size_unit="chars")
    t2 = chunker.hierarchical_chunk("m.py", code, levels=[32, 128], size_unit="chars")
    assert [n.chunk.text for n in t1.nodes] == [n.chunk.text for n in t2.nodes]
    assert [n.level for n in t1.nodes] == [n.level for n in t2.nodes]


def test_hierarchical_chunk_empty_content() -> None:
    chunker = Chunker()
    tree = chunker.hierarchical_chunk("m.py", "", levels=[32, 128], size_unit="chars")
    assert tree.leaves() == []



"""
Code chunking: Python + TypeScript samples, overlap and context_mode,
and byte-range reconstruction checks.
"""

from __future__ import annotations

from omnichunk import Chunker


def main() -> None:
    py_src = '''\
class Greeter:
    """Doc."""

    def hello(self, name: str) -> str:
        return f"hi {name}"

    def bye(self) -> None:
        pass
'''

    ts_src = '''\
interface User { id: number; name: string; }

export function greet(u: User): string {
  return `Hello ${u.name}`;
}
'''

    chunker = Chunker(
        max_chunk_size=120,
        size_unit="chars",
        overlap=20,
        overlap_lines=0,
        context_mode="full",
        sibling_detail="names",
    )
    py_chunks = chunker.chunk("sample.py", py_src)
    print(f"Python chunks: {len(py_chunks)}")
    raw_py = py_src.encode("utf-8")
    for c in py_chunks:
        sl = raw_py[c.byte_range.start : c.byte_range.end]
        assert sl.decode("utf-8") == c.text, "reconstruction failed"
        ents = [e.type.value for e in c.context.entities]
        print(f"  idx={c.index} entities={ents} breadcrumb={c.context.breadcrumb}")
        if c.context.siblings:
            print(f"    siblings: {[s.name for s in c.context.siblings]}")

    chunker_min = Chunker(
        max_chunk_size=120,
        size_unit="chars",
        context_mode="minimal",
    )
    py_min = chunker_min.chunk("sample.py", py_src)
    print(f"Python chunks (minimal context): {len(py_min)}")
    print("  first contextualized length:", len(py_min[0].contextualized_text) if py_min else 0)

    ts_chunks = chunker.chunk("sample.ts", ts_src)
    print(f"TypeScript chunks: {len(ts_chunks)}")
    raw_ts = ts_src.encode("utf-8")
    for c in ts_chunks:
        assert raw_ts[c.byte_range.start : c.byte_range.end].decode("utf-8") == c.text


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path

ROOT = Path(__file__).resolve().parents[1]
FIXTURES = ROOT / "tests" / "fixtures"

_DEFAULT_FIXTURE_FILES: tuple[str, ...] = (
    "python_complex.py",
    "typescript_complex.ts",
    "rust_complex.rs",
    "markdown_doc.md",
    "html_page.html",
    "sample.json",
    "sample.yaml",
    "sample.toml",
    "mixed_notebook.py",
)


@dataclass(frozen=True)
class CorpusEntry:
    filepath: str
    text: str


def collect_corpus_entries(
    *,
    mode: str,
    repeat: int,
    directory: str | None,
    glob_pattern: str,
    max_files: int,
    encoding: str = "utf-8",
) -> list[CorpusEntry]:
    return list(
        iter_corpus_entries(
            mode=mode,
            repeat=repeat,
            directory=directory,
            glob_pattern=glob_pattern,
            max_files=max_files,
            encoding=encoding,
        )
    )


def iter_corpus_entries(
    *,
    mode: str,
    repeat: int,
    directory: str | None,
    glob_pattern: str,
    max_files: int,
    encoding: str = "utf-8",
) -> Iterator[CorpusEntry]:
    if repeat < 1:
        raise ValueError("repeat must be >= 1")
    if max_files < 1:
        raise ValueError("max_files must be >= 1")

    if mode == "fixtures":
        templates = _load_fixture_templates(encoding=encoding)
        for rep in range(repeat):
            for name, text in templates:
                yield CorpusEntry(filepath=f"fixtures/{rep:04d}_{name}", text=text)
        return

    if mode == "mega-python":
        base = (FIXTURES / "python_complex.py").read_text(encoding=encoding)
        parts: list[str] = ["# synthetic large python corpus\n\n"]
        for idx in range(repeat):
            parts.append(f"# block {idx}\n")
            parts.append(base)
            parts.append("\n\n")
        yield CorpusEntry(filepath="synthetic/mega_python.py", text="".join(parts))
        return

    if mode == "directory":
        if directory is None:
            raise ValueError("directory mode requires --directory")

        root = Path(directory)
        if not root.exists():
            raise FileNotFoundError(f"Directory does not exist: {directory}")
        if not root.is_dir():
            raise NotADirectoryError(f"Expected directory but got: {directory}")

        count = 0
        for path in sorted(root.glob(glob_pattern), key=lambda item: item.as_posix()):
            if count >= max_files:
                break
            if not path.is_file():
                continue

            try:
                text = path.read_text(encoding=encoding)
            except (UnicodeDecodeError, OSError):
                continue

            yield CorpusEntry(filepath=str(path), text=text)
            count += 1
        return

    raise ValueError(f"Unsupported corpus mode: {mode}")


def _load_fixture_templates(*, encoding: str) -> list[tuple[str, str]]:
    templates: list[tuple[str, str]] = []
    for name in _DEFAULT_FIXTURE_FILES:
        path = FIXTURES / name
        if not path.exists():
            continue
        templates.append((name, path.read_text(encoding=encoding)))

    if not templates:
        raise FileNotFoundError("No benchmark fixtures were found under tests/fixtures")

    return templates



from __future__ import annotations

from omnichunk.parser.query_patterns import get_query_source

_LANGUAGES_WITH_NAME_CAPTURES = [
    "python",
    "javascript",
    "typescript",
    "rust",
    "go",
    "java",
    "c",
    "cpp",
    "csharp",
    "ruby",
    "php",
    "kotlin",
    "swift",
]


def test_query_source_exists_for_python() -> None:
    source = get_query_source("python")
    assert source is not None
    assert "function_definition" in source
    assert "class_definition" in source
    assert "@name" in source


def test_query_source_missing_for_plaintext() -> None:
    assert get_query_source("plaintext") is None


def test_query_source_exists_for_extended_languages() -> None:
    for language in ["c", "cpp", "csharp", "ruby", "php", "kotlin", "swift"]:
        source = get_query_source(language)  # type: ignore[arg-type]
        assert source is not None
        assert "@entity" in source


def test_query_sources_include_name_captures_for_supported_languages() -> None:
    for language in _LANGUAGES_WITH_NAME_CAPTURES:
        source = get_query_source(language)  # type: ignore[arg-type]
        assert source is not None
        assert "@name" in source



from __future__ import annotations

from pathlib import Path
import sys

FILES = [
    "AI_RULES.md",
    ".cursorrules",
    ".windsurfrules",
    "CLAUDE.md",
    ".github/copilot-instructions.md",
]


def main() -> int:
    root = Path(__file__).resolve().parents[1]
    contents: dict[str, str] = {}

    missing: list[str] = []
    for rel in FILES:
        path = root / rel
        if not path.exists():
            missing.append(rel)
            continue
        contents[rel] = path.read_text(encoding="utf-8")

    if missing:
        print("Missing rule file(s):")
        for rel in missing:
            print(f"- {rel}")
        return 1

    baseline_name = FILES[0]
    baseline = contents[baseline_name]

    mismatched: list[str] = []
    for rel in FILES[1:]:
        if contents[rel] != baseline:
            mismatched.append(rel)

    if mismatched:
        print("Rule files are not synchronized.")
        print(f"Baseline: {baseline_name}")
        print("Mismatched:")
        for rel in mismatched:
            print(f"- {rel}")
        return 1

    print("Rule files are synchronized.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from collections.abc import Callable
from functools import lru_cache
from typing import Literal

from .tokenizers import resolve_tokenizer


def make_token_counter(
    tokenizer: str | Callable[[str], int] | object | None,
    *,
    max_token_chars: int | None = None,
    chunk_size: int | None = None,
    cache_maxsize: int | None = 16384,
) -> Callable[[str], int]:
    """Create a memoized token counter with optional long-text short-circuit."""
    token_counter = resolve_tokenizer(tokenizer)

    @lru_cache(maxsize=cache_maxsize)
    def _memoized_count(text: str) -> int:
        if not text:
            return 0

        if (
            max_token_chars is not None
            and chunk_size is not None
            and chunk_size > 0
            and len(text) > chunk_size * 6
        ):
            probe_len = min(len(text), chunk_size * 6 + max_token_chars)
            probe_count = token_counter(text[:probe_len])
            if probe_count > chunk_size:
                return chunk_size + 1

        return token_counter(text)

    return _memoized_count


def make_size_counter(
    size_unit: Literal["tokens", "chars", "nws"],
    tokenizer: str | Callable[[str], int] | object | None = None,
    *,
    max_token_chars: int | None = None,
    chunk_size: int | None = None,
    cache_maxsize: int | None = 16384,
) -> Callable[[str], int]:
    """Create a generic chunk size counter."""
    if size_unit == "chars":
        return len
    if size_unit == "nws":
        return lambda text: sum(1 for ch in text if not ch.isspace())
    return make_token_counter(
        tokenizer,
        max_token_chars=max_token_chars,
        chunk_size=chunk_size,
        cache_maxsize=cache_maxsize,
    )



from __future__ import annotations

import csv
import hashlib
import importlib
import json
from collections.abc import Sequence
from dataclasses import fields, is_dataclass
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import Any, cast

from omnichunk.types import ByteRange, Chunk, ChunkContext, ContentType, Language, LineRange


def chunk_to_dict(chunk: Chunk) -> dict[str, Any]:
    """Convert a Chunk dataclass into a serializable dictionary."""
    value = _to_serializable(chunk)
    if not isinstance(value, dict):
        return {}
    return value


def chunk_from_dict(data: dict[str, Any]) -> Chunk:
    """Rebuild a :class:`Chunk` from :func:`chunk_to_dict` output (minimal context parse)."""
    ctx_raw = data.get("context") or {}
    ct_raw = ctx_raw.get("content_type", "prose")
    if isinstance(ct_raw, str):
        try:
            content_type = ContentType(ct_raw)
        except ValueError:
            content_type = ContentType.PROSE
    else:
        content_type = ContentType.PROSE

    br = data.get("byte_range") or {}
    lr = data.get("line_range") or {}
    return Chunk(
        text=str(data.get("text", "")),
        contextualized_text=str(data.get("contextualized_text", data.get("text", ""))),
        byte_range=ByteRange(start=int(br.get("start", 0)), end=int(br.get("end", 0))),
        line_range=LineRange(start=int(lr.get("start", 0)), end=int(lr.get("end", 0))),
        index=int(data.get("index", 0)),
        total_chunks=int(data.get("total_chunks", -1)),
        context=ChunkContext(
            filepath=str(ctx_raw.get("filepath", "")),
            language=cast(Language, ctx_raw.get("language", "plaintext")),
            content_type=content_type,
            format_metadata=dict(ctx_raw.get("format_metadata", {})),
        ),
        token_count=int(data.get("token_count", 0)),
        char_count=int(data.get("char_count", 0)),
        nws_count=int(data.get("nws_count", 0)),
    )


def chunks_to_jsonl(chunks: Sequence[Chunk], *, output_path: str | None = None) -> str:
    """Serialize chunks into JSONL. Optionally write output to a file path."""
    lines = [
        json.dumps(chunk_to_dict(chunk), ensure_ascii=False, sort_keys=True)
        for chunk in chunks
    ]
    payload = "\n".join(lines)
    if lines:
        payload += "\n"

    if output_path:
        Path(output_path).write_text(payload, encoding="utf-8")

    return payload


def chunks_to_csv(chunks: Sequence[Chunk], *, output_path: str | None = None) -> str:
    """Serialize chunks into CSV with flattened context fields."""
    columns = [
        "filepath",
        "language",
        "content_type",
        "index",
        "total_chunks",
        "byte_start",
        "byte_end",
        "line_start",
        "line_end",
        "token_count",
        "char_count",
        "nws_count",
        "breadcrumb",
        "section_type",
        "entity_count",
        "text",
        "contextualized_text",
    ]

    buffer = StringIO()
    writer = csv.DictWriter(buffer, fieldnames=columns)
    writer.writeheader()

    for chunk in chunks:
        writer.writerow(
            {
                "filepath": chunk.context.filepath,
                "language": chunk.context.language,
                "content_type": chunk.context.content_type.value,
                "index": chunk.index,
                "total_chunks": chunk.total_chunks,
                "byte_start": chunk.byte_range.start,
                "byte_end": chunk.byte_range.end,
                "line_start": chunk.line_range.start,
                "line_end": chunk.line_range.end,
                "token_count": chunk.token_count,
                "char_count": chunk.char_count,
                "nws_count": chunk.nws_count,
                "breadcrumb": " > ".join(chunk.context.breadcrumb),
                "section_type": chunk.context.section_type,
                "entity_count": len(chunk.context.entities),
                "text": chunk.text,
                "contextualized_text": chunk.contextualized_text,
            }
        )

    payload = buffer.getvalue()

    if output_path:
        Path(output_path).write_text(payload, encoding="utf-8")

    return payload


def chunks_to_langchain_docs(
    chunks: Sequence[Chunk],
    *,
    use_contextualized_text: bool = True,
) -> list[Any]:
    """Convert chunks to LangChain Document objects."""
    document_cls = _import_langchain_document_class()
    docs: list[Any] = []

    for chunk in chunks:
        content = chunk.contextualized_text if use_contextualized_text else chunk.text
        docs.append(
            document_cls(
                page_content=content,
                metadata=_chunk_metadata(chunk),
            )
        )

    return docs


def chunks_to_pinecone_vectors(
    chunks: Sequence[Chunk],
    embeddings: Sequence[list[float]],
    *,
    namespace: str = "",
    use_contextualized_text: bool = True,
) -> list[dict[str, Any]]:
    """Convert chunks to Pinecone upsert-ready dicts (no Pinecone client required).

    Each dict has ``id``, ``values``, ``metadata``. If ``namespace`` is non-empty, a
    top-level ``namespace`` key is set on each dict.
    """
    if len(embeddings) != len(chunks):
        raise ValueError(
            f"embeddings length ({len(embeddings)}) must match chunks length ({len(chunks)})"
        )

    out: list[dict[str, Any]] = []
    for chunk, values in zip(chunks, embeddings):
        row: dict[str, Any] = {
            "id": _stable_vector_id(chunk),
            "values": list(values),
            "metadata": _vectordb_metadata(chunk, use_contextualized_text=use_contextualized_text),
        }
        if namespace:
            row["namespace"] = namespace
        out.append(row)
    return out


def chunks_to_weaviate_objects(
    chunks: Sequence[Chunk],
    embeddings: Sequence[list[float]],
    *,
    class_name: str = "OmnichunkDocument",
    use_contextualized_text: bool = True,
) -> list[dict[str, Any]]:
    """Convert chunks to Weaviate batch-import-ready dicts (no Weaviate client required)."""
    if len(embeddings) != len(chunks):
        raise ValueError(
            f"embeddings length ({len(embeddings)}) must match chunks length ({len(chunks)})"
        )

    return [
        {
            "class": class_name,
            "vector": list(values),
            "properties": _vectordb_metadata(
                chunk,
                use_contextualized_text=use_contextualized_text,
            ),
        }
        for chunk, values in zip(chunks, embeddings)
    ]


def chunks_to_supabase_rows(
    chunks: Sequence[Chunk],
    embeddings: Sequence[list[float]],
    *,
    use_contextualized_text: bool = True,
) -> list[dict[str, Any]]:
    """Convert chunks to Supabase / pgvector INSERT-ready dicts (no client required)."""
    if len(embeddings) != len(chunks):
        raise ValueError(
            f"embeddings length ({len(embeddings)}) must match chunks length ({len(chunks)})"
        )

    rows: list[dict[str, Any]] = []
    for chunk, values in zip(chunks, embeddings):
        meta = _vectordb_metadata(chunk, use_contextualized_text=use_contextualized_text)
        content = meta.pop("text")
        rows.append(
            {
                "content": content,
                "embedding": list(values),
                **meta,
            }
        )
    return rows


def chunks_to_llamaindex_docs(
    chunks: Sequence[Chunk],
    *,
    use_contextualized_text: bool = True,
) -> list[Any]:
    """Convert chunks to LlamaIndex Document objects."""
    document_cls = _import_llamaindex_document_class()
    docs: list[Any] = []

    for chunk in chunks:
        content = chunk.contextualized_text if use_contextualized_text else chunk.text
        docs.append(document_cls(text=content, metadata=_chunk_metadata(chunk)))

    return docs


def stable_chunk_id(chunk: Chunk, filepath: str | None = None) -> str:
    """Return a stable SHA-256 hex ID for a chunk (same as Pinecone/Weaviate export IDs).

    Uses filepath (from ``chunk.context`` or override), chunk index, and byte range.
    Identical to :func:`chunks_to_pinecone_vectors` row ``id`` values.
    """
    fp = filepath if filepath is not None else chunk.context.filepath
    raw = f"{fp}\0{chunk.index}\0{chunk.byte_range.start}\0{chunk.byte_range.end}"
    return hashlib.sha256(raw.encode("utf-8")).hexdigest()


def _stable_vector_id(chunk: Chunk) -> str:
    return stable_chunk_id(chunk)


def _vectordb_metadata(chunk: Chunk, *, use_contextualized_text: bool) -> dict[str, Any]:
    text = chunk.contextualized_text if use_contextualized_text else chunk.text
    return {
        "text": text,
        "filepath": chunk.context.filepath,
        "language": chunk.context.language,
        "content_type": chunk.context.content_type.value,
        "index": chunk.index,
        "total_chunks": chunk.total_chunks,
        "byte_start": chunk.byte_range.start,
        "byte_end": chunk.byte_range.end,
        "line_start": chunk.line_range.start,
        "line_end": chunk.line_range.end,
        "token_count": chunk.token_count,
        "char_count": chunk.char_count,
        "nws_count": chunk.nws_count,
        "breadcrumb": " > ".join(chunk.context.breadcrumb),
    }


def _chunk_metadata(chunk: Chunk) -> dict[str, Any]:
    return {
        "filepath": chunk.context.filepath,
        "language": chunk.context.language,
        "content_type": chunk.context.content_type.value,
        "index": chunk.index,
        "total_chunks": chunk.total_chunks,
        "byte_start": chunk.byte_range.start,
        "byte_end": chunk.byte_range.end,
        "line_start": chunk.line_range.start,
        "line_end": chunk.line_range.end,
        "token_count": chunk.token_count,
        "char_count": chunk.char_count,
        "nws_count": chunk.nws_count,
        "breadcrumb": list(chunk.context.breadcrumb),
        "section_type": chunk.context.section_type,
        "heading_hierarchy": list(chunk.context.heading_hierarchy),
    }


def _import_langchain_document_class() -> Any:
    try:
        documents_module = importlib.import_module("langchain_core.documents")
        document_cls = getattr(documents_module, "Document", None)
        if document_cls is None:
            raise AttributeError("Document type not found in langchain_core.documents")
        return document_cls
    except Exception:
        pass

    try:
        schema_module = importlib.import_module("langchain.schema")
        document_cls = getattr(schema_module, "Document", None)
        if document_cls is None:
            raise AttributeError("Document type not found in langchain.schema")
        return document_cls
    except Exception as exc:
        raise ImportError(
            "LangChain support requires 'langchain-core' or 'langchain' to be installed."
        ) from exc


def _import_llamaindex_document_class() -> Any:
    try:
        schema_module = importlib.import_module("llama_index.core.schema")
        document_cls = getattr(schema_module, "Document", None)
        if document_cls is None:
            raise AttributeError("Document type not found in llama_index.core.schema")
        return document_cls
    except Exception as exc:
        raise ImportError(
            "LlamaIndex support requires 'llama-index-core' (or 'llama-index') to be installed."
        ) from exc


def _to_serializable(value: Any) -> Any:
    if isinstance(value, Enum):
        return value.value

    if is_dataclass(value):
        return {
            field.name: _to_serializable(getattr(value, field.name))
            for field in fields(value)
        }

    if isinstance(value, dict):
        return {str(key): _to_serializable(item) for key, item in value.items()}

    if isinstance(value, list):
        return [_to_serializable(item) for item in value]

    if isinstance(value, tuple):
        return [_to_serializable(item) for item in value]

    return value



"""
Incremental chunk diff engine — stable ID set operations, O(N + M).
"""

from __future__ import annotations

from collections.abc import Sequence

from omnichunk.chunker import Chunker
from omnichunk.serialization import stable_chunk_id
from omnichunk.types import Chunk, ChunkDiff


def chunk_diff(
    filepath: str,
    new_content: str,
    *,
    previous_chunks: Sequence[Chunk],
    chunker: Chunker | None = None,
    **chunker_options: object,
) -> ChunkDiff:
    if chunker is None:
        chunker = Chunker(**chunker_options)

    new_chunks = chunker.chunk(filepath, new_content)

    prev_ids: set[str] = {stable_chunk_id(c) for c in previous_chunks}
    new_ids: set[str] = set()
    added: list[Chunk] = []
    unchanged: list[Chunk] = []

    for c in new_chunks:
        cid = stable_chunk_id(c)
        new_ids.add(cid)
        if cid in prev_ids:
            unchanged.append(c)
        else:
            added.append(c)

    removed_ids = sorted(prev_ids - new_ids)

    return ChunkDiff(added=added, removed_ids=removed_ids, unchanged=unchanged)



from __future__ import annotations

import re
from collections.abc import Callable

_SENT_SPLIT = re.compile(r"(?<=[.!?])\s+(?=[A-Z\"'(\[])")
_PARA_SPLIT = re.compile(r"\n\n+")


def _collapse_empty_triples(
    triples: list[tuple[str, int, int]],
    *,
    full_text: str,
) -> list[tuple[str, int, int]]:
    if not triples:
        return [("", 0, len(full_text))] if full_text == "" else []
    merged: list[tuple[str, int, int]] = []
    for s, a, b in triples:
        if s == "":
            if merged:
                ps, pa, _ = merged[-1]
                merged[-1] = (ps + s, pa, b)
            else:
                merged.append((s, a, b))
        else:
            if merged and merged[-1][0] == "":
                zs, za, _ = merged.pop()
                merged.append((zs + s, za, b))
            else:
                merged.append((s, a, b))
    joined = "".join(s for s, _, _ in merged)
    if joined != full_text:
        raise ValueError("sentence split failed reconstruction invariant")
    return merged


def _default_split_triples(text: str) -> list[tuple[str, int, int]]:
    if not text:
        return [("", 0, 0)]
    parts: list[tuple[str, int, int]] = []
    pos = 0
    n = len(text)
    while pos < n:
        m_para = _PARA_SPLIT.search(text, pos)
        m_sent = _SENT_SPLIT.search(text, pos)
        p_start = m_para.start() if m_para else n + 1
        s_start = m_sent.start() if m_sent else n + 1
        if m_para is not None and p_start <= s_start:
            end = m_para.end()
            parts.append((text[pos:end], pos, end))
            pos = end
        elif m_sent is not None:
            end = m_sent.end()
            parts.append((text[pos:end], pos, end))
            pos = m_sent.end()
        else:
            parts.append((text[pos:], pos, n))
            pos = n
    return _collapse_empty_triples(parts, full_text=text)


def split_sentences(
    text: str,
    *,
    splitter_fn: Callable[[str], list[str]] | None = None,
) -> list[tuple[str, int, int]]:
    """Split text into sentences, returning (sentence_text, char_start, char_end) triples.

    If splitter_fn is provided, it is called with the full text and must return
    a list of sentence strings. The function then maps them back to byte offsets
    using a linear scan (no re-encode; O(N) over chars).

    Default splitter: regex-based, splits on sentence-ending punctuation
    (.!?) followed by whitespace + capital letter OR end of string.
    Does NOT depend on nltk.

    If nltk is installed and splitter_fn is None, the caller may optionally
    pass nltk.tokenize.sent_tokenize as splitter_fn for better accuracy.

    Returns list of (sentence_text, char_start, char_end) where
    char_start is inclusive, char_end is exclusive.
    All returned ranges must cover the full text contiguously.
    Empty strings are merged into the previous sentence, or the next if first.
    """
    if splitter_fn is None:
        return _default_split_triples(text)

    strings = splitter_fn(text)
    out: list[tuple[str, int, int]] = []
    cursor = 0
    for s in strings:
        if s == "":
            if out:
                ps, a, b = out[-1]
                out[-1] = (ps + s, a, b)
            else:
                out.append((s, cursor, cursor))
            continue
        idx = text.find(s, cursor)
        if idx < 0:
            raise ValueError(
                "splitter_fn produced a sentence not found in order in the source text"
            )
        end = idx + len(s)
        out.append((s, idx, end))
        cursor = end
    if cursor != len(text):
        raise ValueError("splitter_fn sentences do not cover the full text")
    return _collapse_empty_triples(out, full_text=text)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from omnichunk.types import Language

from .languages import get_parser as get_ts_parser


@dataclass(frozen=True)
class ParseResult:
    tree: Any | None
    errors: list[str] = field(default_factory=list)


def parse_code(
    code: str,
    language: Language,
    *,
    filepath: str = "",
) -> ParseResult:
    """Parse source code with an optional plugin parser, then tree-sitter when available."""
    from omnichunk import plugins as _plugins

    custom = _plugins.get_parser(str(language))
    if custom is not None:
        try:
            result = custom(filepath, code)
            if result is not None:
                errors: list[str] = []
                root = getattr(result, "root_node", None)
                if root is not None:
                    has_error = getattr(root, "has_error", False)
                    if has_error:
                        errors.append("Tree contains syntax error nodes")
                return ParseResult(tree=result, errors=errors)
        except Exception as exc:
            return ParseResult(tree=None, errors=[f"Plugin parser failed: {exc}"])

    parser = get_ts_parser(language)
    if parser is None:
        return ParseResult(tree=None, errors=[f"No tree-sitter parser available for '{language}'"])

    raw_bytes = code.encode("utf-8")
    try:
        tree = parser.parse(raw_bytes)
    except Exception as exc:
        return ParseResult(tree=None, errors=[f"Tree-sitter parse failed: {exc}"])

    errors_ts: list[str] = []
    root = getattr(tree, "root_node", None)
    if root is not None:
        has_error = getattr(root, "has_error", False)
        if has_error:
            errors_ts.append("Tree contains syntax error nodes")
    return ParseResult(tree=tree, errors=errors_ts)



"""
Heuristic proposition extraction — regex over sentences, zero extra dependencies.
"""

from __future__ import annotations

import re
from typing import Any

from omnichunk.propositions.types import Proposition
from omnichunk.semantic.sentences import split_sentences
from omnichunk.types import ByteRange
from omnichunk.util.text_index import TextIndex

# Patterns: (name, regex, base_confidence)
_PATTERNS: tuple[tuple[str, re.Pattern[str], float], ...] = (
    (
        "copula_is",
        re.compile(
            r"\b([A-Z][a-z]+(?:\s+[a-z]+){0,6})\s+is\s+([^\.!?]+(?:\.|$))",
            re.MULTILINE,
        ),
        0.72,
    ),
    (
        "defined_as",
        re.compile(
            r"\b([A-Z][^\n:]{1,64})\s*:\s*([^\n]+)",
            re.MULTILINE,
        ),
        0.65,
    ),
    (
        "numeric_claim",
        re.compile(
            r"\b(?:approximately|about|over|under|at least|more than|less than)\s+"
            r"[\d,]+(?:\.\d+)?%?\b[^\.!?]*[\.!?]",
            re.IGNORECASE,
        ),
        0.58,
    ),
    (
        "year_claim",
        re.compile(r"\b(?:in|since|until|before|after)\s+(?:the\s+)?(?:year\s+)?\d{4}\b[^\.!?]*[\.!?]"),
        0.55,
    ),
)


def extract_propositions_heuristic(filepath: str, text: str) -> list[Proposition]:
    """Extract propositions using sentence split + lightweight regex patterns."""
    if not text.strip():
        return []

    ti = TextIndex(text)
    sents = split_sentences(text)
    out: list[Proposition] = []
    seen: set[tuple[int, int]] = set()

    for sent, c0, _c1 in sents:
        if not sent.strip():
            continue
        for pname, rx, base in _PATTERNS:
            for m in rx.finditer(sent):
                a, b = m.span()
                abs_start = c0 + a
                abs_end = c0 + b
                span = (abs_start, abs_end)
                if span in seen:
                    continue
                seen.add(span)
                snippet = text[abs_start:abs_end].strip()
                if not snippet:
                    continue
                bs = ti.byte_offset_for_char(abs_start)
                be = ti.byte_offset_for_char(abs_end)
                meta: dict[str, Any] = {
                    "pattern": pname,
                    "filepath": filepath,
                }
                conf = min(0.95, base + 0.03 * min(len(snippet), 120) / 120.0)
                out.append(
                    Proposition(
                        text=snippet,
                        byte_range=ByteRange(start=bs, end=be),
                        confidence=round(conf, 4),
                        metadata=meta,
                    )
                )

    out.sort(key=lambda p: (p.byte_range.start, p.byte_range.end, p.text))
    return out



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Literal


class ContentType(Enum):
    CODE = "code"
    PROSE = "prose"
    MARKUP = "markup"
    HYBRID = "hybrid"


class EntityType(Enum):
    FUNCTION = "function"
    METHOD = "method"
    CLASS = "class"
    INTERFACE = "interface"
    TYPE = "type"
    ENUM = "enum"
    IMPORT = "import"
    EXPORT = "export"
    MODULE = "module"
    DECORATOR = "decorator"
    CONSTANT = "constant"
    HEADING = "heading"
    SECTION = "section"
    PARAGRAPH = "paragraph"
    LIST = "list"
    TABLE = "table"
    CODE_BLOCK = "code_block"
    FRONTMATTER = "frontmatter"


Language = Literal[
    "python",
    "javascript",
    "typescript",
    "rust",
    "go",
    "java",
    "c",
    "cpp",
    "csharp",
    "ruby",
    "php",
    "swift",
    "kotlin",
    "scala",
    "haskell",
    "lua",
    "zig",
    "elixir",
    "markdown",
    "html",
    "xml",
    "yaml",
    "json",
    "toml",
    "plaintext",
    "latex",
    "jupyter",
    "pdf",
    "docx",
]


@dataclass(frozen=True)
class ByteRange:
    start: int
    end: int


@dataclass(frozen=True)
class LineRange:
    start: int
    end: int


@dataclass(frozen=True)
class EntityInfo:
    """An entity detected within source content."""

    name: str
    type: EntityType
    signature: str = ""
    docstring: str | None = None
    byte_range: ByteRange | None = None
    line_range: LineRange | None = None
    is_partial: bool = False
    parent: str | None = None


@dataclass(frozen=True)
class SiblingInfo:
    """A sibling entity near the current chunk."""

    name: str
    type: EntityType
    position: Literal["before", "after"]
    distance: int
    signature: str = ""


@dataclass(frozen=True)
class ImportInfo:
    """An import dependency."""

    name: str
    source: str
    is_default: bool = False
    is_namespace: bool = False


@dataclass(frozen=True)
class ChunkContext:
    """Rich contextual metadata for a chunk."""

    filepath: str = ""
    language: Language = "plaintext"
    content_type: ContentType = ContentType.PROSE

    scope: list[EntityInfo] = field(default_factory=list)
    breadcrumb: list[str] = field(default_factory=list)
    entities: list[EntityInfo] = field(default_factory=list)
    siblings: list[SiblingInfo] = field(default_factory=list)
    imports: list[ImportInfo] = field(default_factory=list)

    heading_hierarchy: list[str] = field(default_factory=list)
    section_type: str = ""
    parse_errors: list[str] = field(default_factory=list)
    format_metadata: dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True)
class Chunk:
    """A single chunk with full context."""

    text: str
    contextualized_text: str

    byte_range: ByteRange
    line_range: LineRange
    index: int
    total_chunks: int

    context: ChunkContext

    token_count: int = 0
    char_count: int = 0
    nws_count: int = 0


@dataclass(frozen=True)
class ChunkNode:
    """A single node in a ChunkTree — wraps a Chunk with parent/child links."""

    chunk: Chunk
    level: int
    """0 = finest granularity (leaves), max_level = coarsest (roots)."""
    parent_index: int | None
    """Index into ChunkTree.nodes of this node's parent, or None if root."""
    child_indices: tuple[int, ...]
    """Sorted indices into ChunkTree.nodes of this node's children."""


@dataclass
class ChunkTree:
    """Multi-level chunk hierarchy produced by hierarchical_chunk()."""

    nodes: list[ChunkNode]
    level_count: int

    def leaves(self) -> list[Chunk]:
        """Return all level-0 chunks in source order."""
        return [n.chunk for n in self.nodes if n.level == 0]

    def roots(self) -> list[Chunk]:
        """Return all top-level (coarsest) chunks in source order."""
        max_level = self.level_count - 1
        return [n.chunk for n in self.nodes if n.level == max_level]

    def at_level(self, level: int) -> list[Chunk]:
        """Return all chunks at a given level in source order."""
        if level < 0 or level >= self.level_count:
            raise ValueError(f"level must be in [0, {self.level_count - 1}]")
        return [n.chunk for n in self.nodes if n.level == level]

    def parent(self, chunk: Chunk) -> Chunk | None:
        """Return the parent chunk of the given chunk, or None if root."""
        for node in self.nodes:
            if node.chunk is chunk and node.parent_index is not None:
                return self.nodes[node.parent_index].chunk
        return None

    def children(self, chunk: Chunk) -> list[Chunk]:
        """Return the children of the given chunk, in source order."""
        for node in self.nodes:
            if node.chunk is chunk:
                ordered = sorted(
                    node.child_indices,
                    key=lambda i: self.nodes[i].chunk.byte_range.start,
                )
                return [self.nodes[i].chunk for i in ordered]
        return []

    def to_flat_list(self, level: int | None = None) -> list[Chunk]:
        """Return all chunks at `level`, or all chunks if level is None."""
        if level is None:
            return [n.chunk for n in self.nodes]
        return self.at_level(level)

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a JSON-compatible dict for storage/transport."""
        from omnichunk.serialization import chunk_to_dict

        return {
            "level_count": self.level_count,
            "nodes": [
                {
                    "level": n.level,
                    "parent_index": n.parent_index,
                    "child_indices": list(n.child_indices),
                    "chunk": chunk_to_dict(n.chunk),
                }
                for n in self.nodes
            ],
        }


@dataclass(frozen=True)
class ChunkDiff:
    """Result of incremental re-chunking after a content change."""

    added: list[Chunk]
    removed_ids: list[str]
    unchanged: list[Chunk]

    @property
    def total_added(self) -> int:
        return len(self.added)

    @property
    def total_removed(self) -> int:
        return len(self.removed_ids)

    @property
    def total_unchanged(self) -> int:
        return len(self.unchanged)


@dataclass(frozen=True)
class ChunkQualityScore:
    """Quality metrics for a chunk."""

    index: int
    overall: float
    entity_integrity: float
    scope_consistency: float
    size_balance: float
    size_value: int


@dataclass(frozen=True)
class ChunkStats:
    """Aggregate statistics for a chunk set."""

    total_chunks: int
    average_size: float
    min_size: int
    max_size: int
    size_unit: Literal["tokens", "chars", "nws"]
    entity_distribution: dict[str, int] = field(default_factory=dict)


@dataclass
class ChunkOptions:
    """Configuration for chunking behavior."""

    max_chunk_size: int = 1500
    min_chunk_size: int = 50
    size_unit: Literal["tokens", "chars", "nws"] = "tokens"
    tokenizer: str | Callable[[str], int] | None = None
    nws_backend: Literal["auto", "python", "rust"] = "auto"

    context_mode: Literal["none", "minimal", "full"] = "full"
    sibling_detail: Literal["none", "names", "signatures"] = "signatures"
    max_siblings: int = 3
    include_imports: bool = True
    filter_imports: bool = False

    overlap: int | float | None = None
    overlap_lines: int = 0

    language: Language | None = None
    content_type: ContentType | None = None
    filepath: str = ""

    preserve_decorators: bool = True
    preserve_comments: bool = True
    include_header_in_sections: bool = True

    _precomputed_nws_cumsum: Any | None = None
    _precomputed_text_index: Any | None = None

    # Semantic chunking options
    semantic: bool = False
    semantic_embed_fn: object = None
    semantic_window: int = 3
    semantic_threshold: float = 0.3
    semantic_min_sentences: int = 1
    semantic_sentence_splitter: object = None

    # Optional OpenTelemetry tracer (opentelemetry-api Tracer); None disables spans.
    otel_tracer: object | None = None


@dataclass(frozen=True)
class UpsertBatch:
    """One batch of vector-DB rows produced by :meth:`Chunker.stream_upsert`."""

    adapter: Literal["pinecone", "weaviate", "supabase"]
    rows: list[dict[str, Any]]
    chunks: tuple[Chunk, ...] = ()


@dataclass(frozen=True)
class BatchResult:
    filepath: str
    chunks: list[Chunk] = field(default_factory=list)
    error: str | None = None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from omnichunk.types import ByteRange


@dataclass(frozen=True)
class Proposition:
    """A short, self-contained factual claim extracted from a document."""

    text: str
    byte_range: ByteRange
    confidence: float
    metadata: dict[str, Any] = field(default_factory=dict)



from .languages import get_language, get_parser, is_supported_code_language
from .query_patterns import get_query_source
from .tree_sitter import parse_code

__all__ = [
    "get_language",
    "get_parser",
    "get_query_source",
    "is_supported_code_language",
    "parse_code",
]



# %% [markdown]
"""
# Notebook Title

This notebook demonstrates mixed content.
"""

# %%
import math


def compute(x: float) -> float:
    """Compute something from x.

    This docstring is intentionally long enough to trigger hybrid handling.
    It contains prose-like material and explanation.
    """
    return math.sqrt(x)


# %% [markdown]
"""
## Notes

- Point A
- Point B
"""



from __future__ import annotations

import re
from collections.abc import Sequence

import numpy as np


def _tokenize(text: str) -> list[str]:
    """Simple whitespace + punctuation tokenizer. Returns lowercase tokens."""
    return re.findall(r"[a-zA-Z][a-zA-Z0-9]*", text.lower())


def build_tfidf_matrix(
    documents: Sequence[str],
    *,
    max_vocab: int = 4096,
    min_df: int = 1,
) -> np.ndarray:
    """Build TF-IDF matrix of shape (N, V) for N documents, V vocab terms."""
    n_docs = len(documents)
    if n_docs == 0:
        return np.zeros((0, 1), dtype=np.float64)

    doc_tokens: list[list[str]] = [_tokenize(d) for d in documents]
    df: dict[str, int] = {}
    for tokens in doc_tokens:
        seen: set[str] = set()
        for t in tokens:
            if t not in seen:
                seen.add(t)
                df[t] = df.get(t, 0) + 1

    candidates = sorted(
        ((t, c) for t, c in df.items() if c >= min_df),
        key=lambda x: (-x[1], x[0]),
    )
    vocab = [t for t, _ in candidates[:max_vocab]]
    if not vocab:
        return np.zeros((n_docs, 1), dtype=np.float64)

    t2i = {t: i for i, t in enumerate(vocab)}
    v = len(vocab)
    tf_mat = np.zeros((n_docs, v), dtype=np.float64)
    for i, tokens in enumerate(doc_tokens):
        for tok in tokens:
            j = t2i.get(tok)
            if j is not None:
                tf_mat[i, j] += 1.0

    idf = np.zeros(v, dtype=np.float64)
    nd = float(n_docs)
    for j, term in enumerate(vocab):
        dfi = float(df[term])
        idf[j] = np.log((1.0 + nd) / (1.0 + dfi)) + 1.0

    out = np.zeros((n_docs, v), dtype=np.float64)
    for i in range(n_docs):
        for j in range(v):
            tf = tf_mat[i, j]
            if tf > 0:
                out[i, j] = (1.0 + np.log(tf)) * idf[j]

    norms = np.linalg.norm(out, axis=1, keepdims=True)
    norms = np.where(norms == 0, 1.0, norms)
    out = out / norms
    return out


def detect_topic_shifts(
    sentences: Sequence[str],
    *,
    window: int = 5,
    threshold: float = 0.5,
    min_shift_gap: int = 3,
    max_vocab: int = 2048,
) -> tuple[int, ...]:
    """Detect topic shifts using TF-IDF cosine similarity over sliding windows."""
    n = len(sentences)
    w = int(window)
    if n < 2 * w:
        return ()

    last_shift_after: int | None = None
    out: list[int] = []

    for i in range(w, n - w + 1):
        before = "".join(sentences[i - w : i])
        after = "".join(sentences[i : i + w])
        mat = build_tfidf_matrix([before, after], max_vocab=max_vocab)
        if mat.shape[1] == 0:
            sim = 0.0
        else:
            v0, v1 = mat[0], mat[1]
            denom = float(np.linalg.norm(v0) * np.linalg.norm(v1))
            sim = float(np.dot(v0, v1) / denom) if denom > 0 else 0.0

        boundary_after = i - 1
        if sim < threshold and (
            last_shift_after is None
            or boundary_after - last_shift_after >= min_shift_gap
        ):
            out.append(boundary_after)
            last_shift_after = boundary_after

    return tuple(out)



from __future__ import annotations

from collections.abc import Sequence
from typing import Literal, cast

from omnichunk.types import Chunk, ChunkQualityScore, ChunkStats

SizeUnit = Literal["tokens", "chars", "nws"]


def compute_chunk_quality_scores(
    chunks: Sequence[Chunk],
    *,
    min_chunk_size: int,
    max_chunk_size: int,
    size_unit: str,
) -> list[ChunkQualityScore]:
    """Compute deterministic quality scores for each chunk."""
    if not chunks:
        return []

    lower = max(1, int(min_chunk_size))
    upper = max(lower, int(max_chunk_size))
    unit = _normalize_size_unit(size_unit)

    out: list[ChunkQualityScore] = []
    for chunk in chunks:
        size_value = _chunk_size_value(chunk, unit)
        entity_integrity = _entity_integrity(chunk)
        scope_consistency = _scope_consistency(chunk)
        size_balance = _size_balance(size_value=size_value, min_size=lower, max_size=upper)

        overall = (entity_integrity * 0.45) + (scope_consistency * 0.3) + (size_balance * 0.25)
        out.append(
            ChunkQualityScore(
                index=chunk.index,
                overall=round(_clamp01(overall), 4),
                entity_integrity=round(entity_integrity, 4),
                scope_consistency=round(scope_consistency, 4),
                size_balance=round(size_balance, 4),
                size_value=size_value,
            )
        )

    return out


def compute_chunk_stats(chunks: Sequence[Chunk], *, size_unit: str) -> ChunkStats:
    """Compute aggregate deterministic stats for chunk collections."""
    unit = _normalize_size_unit(size_unit)
    if not chunks:
        return ChunkStats(
            total_chunks=0,
            average_size=0.0,
            min_size=0,
            max_size=0,
            size_unit=unit,
            entity_distribution={},
        )

    sizes = [_chunk_size_value(chunk, unit) for chunk in chunks]
    average = sum(sizes) / len(sizes)

    entity_distribution: dict[str, int] = {}
    for chunk in chunks:
        for entity in chunk.context.entities:
            key = entity.type.value
            entity_distribution[key] = entity_distribution.get(key, 0) + 1

    return ChunkStats(
        total_chunks=len(chunks),
        average_size=round(average, 4),
        min_size=min(sizes),
        max_size=max(sizes),
        size_unit=unit,
        entity_distribution=dict(sorted(entity_distribution.items())),
    )


def _entity_integrity(chunk: Chunk) -> float:
    entities = list(chunk.context.entities)
    if not entities:
        return 1.0

    partial_count = sum(1 for entity in entities if entity.is_partial)
    score = 1.0 - (partial_count / len(entities))
    return _clamp01(score)


def _scope_consistency(chunk: Chunk) -> float:
    scope_labels: set[str] = set()

    if chunk.context.breadcrumb:
        scope_labels.add("::".join(chunk.context.breadcrumb))

    for entity in chunk.context.entities:
        if entity.parent:
            scope_labels.add(entity.parent)
        elif entity.name:
            scope_labels.add(entity.name)

    if not scope_labels:
        return 1.0

    distinct = len(scope_labels)
    if distinct <= 1:
        return 1.0
    if distinct == 2:
        return 0.75

    penalty = 0.22 * (distinct - 2)
    return _clamp01(0.75 - penalty)


def _size_balance(*, size_value: int, min_size: int, max_size: int) -> float:
    if max_size <= min_size:
        return 1.0

    if size_value <= min_size or size_value >= max_size:
        return 0.0

    midpoint = (min_size + max_size) / 2
    half_span = (max_size - min_size) / 2
    if half_span <= 0:
        return 1.0

    distance = abs(size_value - midpoint)
    score = 1.0 - (distance / half_span)
    return _clamp01(score)


def _chunk_size_value(chunk: Chunk, size_unit: str) -> int:
    if size_unit == "tokens":
        return int(chunk.token_count)
    if size_unit == "nws":
        return int(chunk.nws_count)
    return int(chunk.char_count)


def _normalize_size_unit(size_unit: str) -> SizeUnit:
    if size_unit in {"tokens", "chars", "nws"}:
        return cast(SizeUnit, size_unit)
    return "chars"


def _clamp01(value: float) -> float:
    if value < 0:
        return 0.0
    if value > 1:
        return 1.0
    return value



from __future__ import annotations

import json
from pathlib import Path

from omnichunk.cli import main


def test_cli_outputs_jsonl_for_file(tmp_path: Path, capsys) -> None:
    file_path = tmp_path / "sample.py"
    file_path.write_text("def ping():\n    return 'pong'\n", encoding="utf-8")

    exit_code = main(
        [
            str(file_path),
            "--format",
            "jsonl",
            "--size-unit",
            "chars",
            "--max-size",
            "40",
        ]
    )

    captured = capsys.readouterr()
    assert exit_code == 0

    lines = [line for line in captured.out.splitlines() if line.strip()]
    assert lines
    first = json.loads(lines[0])
    assert first["context"]["filepath"] == str(file_path)


def test_cli_stats_for_directory(tmp_path: Path, capsys) -> None:
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    (src_dir / "a.py").write_text("def a():\n    return 1\n", encoding="utf-8")
    (src_dir / "b.py").write_text("def b():\n    return 2\n", encoding="utf-8")

    exit_code = main(
        [
            str(src_dir),
            "--glob",
            "**/*.py",
            "--size-unit",
            "chars",
            "--max-size",
            "48",
            "--stats",
        ]
    )

    captured = capsys.readouterr()
    assert exit_code == 0

    payload = json.loads(captured.out)
    assert payload["total_chunks"] >= 2
    assert payload["size_unit"] == "chars"
    assert payload["failed_files"] == 0


def test_cli_accepts_nws_backend_flag(tmp_path: Path, capsys) -> None:
    file_path = tmp_path / "sample.py"
    file_path.write_text("def ping():\n    return 'pong'\n", encoding="utf-8")

    exit_code = main(
        [
            str(file_path),
            "--format",
            "json",
            "--size-unit",
            "chars",
            "--max-size",
            "40",
            "--nws-backend",
            "python",
        ]
    )

    captured = capsys.readouterr()
    assert exit_code == 0
    payload = json.loads(captured.out)
    assert payload
